#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
import operator
import typing as t
from copy import deepcopy
from enum import Enum
from numbers import Number

import numpy as np
import xarray as xr

from pyxel.evaluator import eval_entry
from pyxel.pipelines import DetectionPipeline, ModelGroup
from pyxel.state import get_obj_att
from pyxel.util.memory import get_size

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


class ResultType(Enum):
    """Result type class."""

    Image = "image"
    Signal = "signal"
    Pixel = "pixel"
    All = "all"


def result_keys(result_type: ResultType = ResultType.All) -> t.Sequence[str]:
    """Return result keys based on result type.

    Parameters
    ----------
    result_type

    Returns
    -------
    list
    """
    if result_type == ResultType.Image:
        return ["image"]
    elif result_type == ResultType.Signal:
        return ["signal"]
    elif result_type == ResultType.Pixel:
        return ["pixel"]
    elif result_type == ResultType.All:
        return ["image", "signal", "pixel"]
    else:
        raise ValueError("Result type unknown.")


# TODO: Is this class needed ?
class Processor:
    """TBW."""

    def __init__(self, detector: "Detector", pipeline: DetectionPipeline):
        self._log = logging.getLogger(__name__)

        self.detector = detector
        self.pipeline = pipeline
        self._result = None  # type: t.Optional[dict]

        self._numbytes = 0

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<detector={self.detector!r}, pipeline={self.pipeline!r}>"

    def __deepcopy__(self, memodict: dict) -> "Processor":
        return Processor(
            detector=deepcopy(self.detector, memo=memodict),
            pipeline=deepcopy(self.pipeline, memo=memodict),
        )

    # TODO: Could it be renamed '__contains__' ?
    # TODO: reimplement this method.
    def has(self, key: str) -> bool:
        """TBW.

        Parameters
        ----------
        key : str

        Returns
        -------
        bool
            TBW.
        """
        found = False
        obj, att = get_obj_att(self, key)
        if isinstance(obj, dict) and att in obj:
            found = True
        elif hasattr(obj, att):
            found = True
        return found

    # TODO: Could it be renamed '__getitem__' ?
    # TODO: Is it really needed ?
    # TODO: What are the valid keys ? (e.g. 'detector.image.array',
    #       'detector.signal.array' and 'detector.pixel.array')
    def get(self, key: str) -> np.ndarray:
        """TBW.

        Parameters
        ----------
        key : str

        Returns
        -------
        TBW.
        """
        # return get_value(self, key)

        func = operator.attrgetter(key)  # type: t.Callable
        result = func(self)

        return np.asarray(result, dtype=float)

    # TODO: Could it be renamed '__setitem__' ?
    def set(
        self,
        key: str,
        value: t.Union[
            str, Number, np.ndarray, t.List[t.Union[str, Number, np.ndarray]]
        ],
        convert_value: bool = True,
    ) -> None:
        """TBW.

        Parameters
        ----------
        key : str
        value
        convert_value : bool
        """
        if convert_value:  # and value:
            # TODO: Refactor this
            # convert the string based value to a number
            if isinstance(value, list):
                new_value_lst = []  # type: t.List[t.Union[str, Number, np.ndarray]]
                for val in value:  # type: t.Union[str, Number, np.ndarray]
                    new_val = eval_entry(val) if val else val
                    new_value_lst.append(new_val)

                new_value = (
                    new_value_lst
                )  # type: t.Union[str, Number, np.ndarray, t.List[t.Union[str, Number, np.ndarray]]]

            else:
                converted_value = eval_entry(
                    value
                )  # type: t.Union[str, Number, np.ndarray]

                new_value = converted_value
        else:
            new_value = value

        obj, att = get_obj_att(self, key)

        if isinstance(obj, dict) and att in obj:
            obj[att] = new_value
        else:
            setattr(obj, att, new_value)

    # TODO: Create a method `DetectionPipeline.run`
    def run_pipeline(self, abort_before: t.Optional[str] = None) -> "Processor":
        """Run a pipeline with all its models in the right order.

        Parameters
        ----------
        abort_before : str
            model name, the pipeline should be aborted before this

        Returns
        -------
        Processor
            TBW.

        Notes
        -----
        The ``Processor`` instance with method '.run_pipeline' is modified.
        """
        self._log.info("Start pipeline")

        # TODO: Use with-statement to set/unset ._is_running
        self.pipeline._is_running = True

        for group_name in self.pipeline.model_group_names:
            # Get a group of models
            models_grp = getattr(
                self.pipeline, group_name
            )  # type: t.Optional[ModelGroup]

            if models_grp:
                self._log.info("Processing group: %r", group_name)

                abort_flag = models_grp.run(
                    detector=self.detector,
                    pipeline=self.pipeline,
                    abort_model=abort_before,
                )
                if abort_flag:
                    break

        self.pipeline._is_running = False

        # TODO: Is is necessary to return 'self' ??
        return self

    @property
    def result(self) -> dict:
        """Return exposure pipeline final result in a dicitonary."""
        if not self._result:
            raise ValueError("No result saved in the processor.")
        return self._result

    @result.setter
    def result(self, result_to_save: dict) -> None:
        """Set result."""
        self._result = result_to_save

    def result_to_dataset(
        self, y: range, x: range, times: np.ndarray, result_type: ResultType
    ) -> xr.Dataset:
        """Return the result in an xarray dataset."""
        if not self._result:
            raise ValueError("No result saved in the processor.")

        lst = []

        for key in result_keys(result_type):

            da = xr.DataArray(
                self.result[key],
                dims=("readout_time", "y", "x"),
                name=key,
                coords={"readout_time": times, "y": y, "x": x},
            )

            lst.append(da)

        return xr.merge(lst)

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to load charge profiles."""

import typing as t
from pathlib import Path

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector, Geometry
from pyxel.util import load_cropped_and_aligned_image


def load_charge(
    detector: Detector,
    filename: t.Union[str, Path],
    position: t.Tuple[int, int] = (0, 0),
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    time_scale: float = 1.0,
) -> None:
    """Load charge from txt file for detector, mostly for but not limited to :term:`CCDs<CCD>`.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    filename : str or Path
        File path.
    position: tuple
        Indices of starting row and column, used when fitting charge to detector.
    align: Literal
        Keyword to align the charge to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    time_scale: float
        Time scale of the input charge, default is 1 second. 0.001 would be ms.
    """
    geo = detector.geometry  # type: Geometry
    position_y, position_x = position

    # Load charge profile as numpy array.
    charges = load_cropped_and_aligned_image(
        shape=(geo.row, geo.col),
        filename=filename,
        position_x=position_x,
        position_y=position_y,
        align=align,
    )  # type: np.ndarray

    charges *= detector.time_step / time_scale

    # Add charges in 'detector'
    detector.charge.add_charge_array(charges)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Amplifier crosstalk model: https://arxiv.org/abs/1808.00790."""

import typing as t
from pathlib import Path

import numba
import numpy as np

from pyxel import load_table

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


# TODO: Put more info in documentation
@numba.njit
def flip_array(array: np.ndarray, direction: int) -> np.ndarray:
    """Flip the array for read direction as in case 1 or back.

    Parameters
    ----------
    array: ndarray
    direction: int

    Returns
    -------
    ndarray
    """
    if direction == 1:
        result = array  # type: np.ndarray
    elif direction == 2:
        result = np.fliplr(array)
    elif direction == 3:
        result = np.flipud(array)
    elif direction == 4:
        result = np.fliplr(np.flipud(array))
    else:
        raise ValueError("Unknown readout direction.")

    return result


@numba.njit
def get_channel_slices(
    shape: tuple, channel_matrix: np.ndarray
) -> t.List[t.List[t.Tuple[t.Any, t.Any]]]:
    """Get pairs of slices that correspond to the given channel matrix in numerical order of channels.

    Parameters
    ----------
    shape: tuple
    channel_matrix: ndarray

    Returns
    -------
    slices
    """
    j_x = channel_matrix.shape[0]
    j_y = 1
    if channel_matrix.ndim == 2:
        j_y = channel_matrix.shape[1]

    delta_x = shape[0] // j_x
    delta_y = shape[1] // j_y

    slices = []

    for j in range(1, channel_matrix.size + 1):
        channel_position = np.argwhere(channel_matrix == j)[0]
        if channel_position.size == 1:
            channel_position = np.append(np.array([0]), channel_position)
        channel_slice_x = (
            channel_position[1] * delta_x,
            (channel_position[1] + 1) * delta_x,
        )
        channel_slice_y = (
            channel_position[0] * delta_y,
            (channel_position[0] + 1) * delta_y,
        )
        slices.append([channel_slice_x, channel_slice_y])

    return slices


def get_matrix(coupling_matrix: t.Union[str, Path, t.Sequence]) -> np.ndarray:
    """Get the coupling matrix either from configuration input or a file.

    Parameters
    ----------
    coupling_matrix : str, Path or sequence of numbers.
        Matrix to create.

    Returns
    -------
    array
        Matrix.
    """
    if isinstance(coupling_matrix, (str, Path)):
        return np.array(load_table(coupling_matrix))
    else:
        return np.array(coupling_matrix)


@numba.njit
def crosstalk_signal_ac(
    array: np.ndarray,
    coupling_matrix: np.ndarray,
    channel_matrix: np.ndarray,
    readout_directions: np.ndarray,
) -> np.ndarray:
    """Apply AC crosstalk signal to an array.

    Parameters
    ----------
    array: ndarray
    coupling_matrix: ndarray
        2D array.
    channel_matrix: ndarray
    readout_directions: ndarray

    Returns
    -------
    ndarray
    """
    amp_number = channel_matrix.size  # number of amplifiers

    slices = get_channel_slices(
        shape=array.shape, channel_matrix=channel_matrix
    )  # type: t.List[t.List[t.Tuple[t.Any, t.Any]]]

    array_copy = array.copy()

    for k in range(amp_number):
        for j in range(amp_number):
            if k != j and coupling_matrix[k][j] != 0:
                s_k = flip_array(
                    array_copy[
                        slices[k][1][0] : slices[k][1][1],
                        slices[k][0][0] : slices[k][0][1],
                    ],
                    readout_directions[k],
                )
                s_k_shift = np.hstack((s_k[:, 0:1], s_k[:, 0:-1]))
                delta_s = s_k - s_k_shift
                crosstalk_signal = coupling_matrix[k][j] * flip_array(
                    delta_s, readout_directions[j]
                )
                array[
                    slices[j][1][0] : slices[j][1][1], slices[j][0][0] : slices[j][0][1]
                ] += crosstalk_signal

    return array


@numba.njit
def crosstalk_signal_dc(
    array: np.ndarray,
    coupling_matrix: np.ndarray,
    channel_matrix: np.ndarray,
    readout_directions: np.ndarray,
) -> np.ndarray:
    """Apply DC crosstalk signal to an array.

    Parameters
    ----------
    array: ndarray
    coupling_matrix: ndarray
    channel_matrix: ndarray
    readout_directions: ndarray

    Returns
    -------
    ndarray
    """
    amp_number = channel_matrix.size  # number of amplifiers

    slices = get_channel_slices(
        shape=array.shape, channel_matrix=channel_matrix
    )  # type: t.List[t.List[t.Tuple[t.Any, t.Any]]]

    array_copy = array.copy()

    for k in range(amp_number):
        for j in range(amp_number):
            if k != j and coupling_matrix[k][j] != 0:
                s_k = flip_array(
                    array_copy[
                        slices[k][1][0] : slices[k][1][1],
                        slices[k][0][0] : slices[k][0][1],
                    ],
                    readout_directions[k],
                )
                crosstalk_signal = coupling_matrix[k][j] * flip_array(
                    s_k, readout_directions[j]
                )
                array[
                    slices[j][1][0] : slices[j][1][1], slices[j][0][0] : slices[j][0][1]
                ] += crosstalk_signal

    return array


def dc_crosstalk(
    detector: "Detector",
    coupling_matrix: t.Union[str, Path, t.Sequence],
    channel_matrix: t.Sequence,
    readout_directions: t.Sequence,
) -> None:
    """Apply DC crosstalk signal to detector signal.

    Parameters
    ----------
    detector: Detector
    coupling_matrix: ndarray
    channel_matrix: ndarray
    readout_directions: ndarray

    Raises
    ------
    ValueError
        If at least one parameter 'coupling_matrix', 'channel_matrix' or
        'readout_directions' does not have the right shape.
    """
    # Validation and conversion
    cpl_matrix_2d = get_matrix(coupling_matrix)  # type: np.ndarray
    ch_matrix = np.array(channel_matrix)  # type: np.ndarray
    directions = np.array(readout_directions)  # type: np.ndarray

    if cpl_matrix_2d.ndim != 2:
        raise ValueError("Expecting 2D 'coupling_matrix'.")

    if cpl_matrix_2d.shape != (ch_matrix.size, ch_matrix.size):
        raise ValueError(
            f"Expecting a matrix of {ch_matrix.size}x{ch_matrix.size} "
            f"elements for 'coupling_matrix'"
        )

    if detector.geometry.row % ch_matrix.shape[0] != 0:
        raise ValueError(
            "Can't split detector array horizontally for a given number of amplifiers."
        )
    if len(ch_matrix.shape) > 1:
        if detector.geometry.col % ch_matrix.shape[1] != 0:
            raise ValueError(
                "Can't split detector array vertically "
                "for a given number of amplifiers."
            )
    if ch_matrix.size != directions.size:
        raise ValueError(
            "Channel matrix and readout directions arrays not the same size."
        )

    # Processing
    signal_2d = crosstalk_signal_dc(
        array=detector.signal.array.copy(),
        coupling_matrix=cpl_matrix_2d,
        channel_matrix=ch_matrix,
        readout_directions=directions,
    )

    detector.signal.array = signal_2d


def ac_crosstalk(
    detector: "Detector",
    coupling_matrix: t.Union[str, Path, t.Sequence],
    channel_matrix: t.Sequence,
    readout_directions: t.Sequence,
) -> None:
    """Apply AC crosstalk signal to detector signal.

    Parameters
    ----------
    detector: Detector
    coupling_matrix: ndarray
    channel_matrix: ndarray
    readout_directions: ndarray

    Raises
    ------
    ValueError
        If at least one parameter 'coupling_matrix', 'channel_matrix' or
        'readout_directions' does not have the right shape.
    """
    # Validation and conversion
    cpl_matrix_2d = get_matrix(coupling_matrix)  # type: np.ndarray
    ch_matrix = np.array(channel_matrix)  # type: np.ndarray
    directions = np.array(readout_directions)  # type: np.ndarray

    if cpl_matrix_2d.ndim != 2:
        raise ValueError("Expecting 2D 'coupling_matrix'.")

    if cpl_matrix_2d.shape != (ch_matrix.size, ch_matrix.size):
        raise ValueError(
            f"Expecting a matrix of {ch_matrix.size}x{ch_matrix.size} "
            f"elements for 'coupling_matrix'"
        )

    if detector.geometry.row % ch_matrix.shape[0] != 0:
        raise ValueError(
            "Can't split detector array horizontally for a given number of amplifiers."
        )
    if len(ch_matrix.shape) > 1:
        if detector.geometry.col % ch_matrix.shape[1] != 0:
            raise ValueError(
                "Can't split detector array vertically "
                "for a given number of amplifiers."
            )
    if ch_matrix.size != directions.size:
        raise ValueError(
            "Channel matrix and readout directions arrays not the same size."
        )

    # Processing
    signal_2d = crosstalk_signal_ac(
        array=detector.signal.array.copy(),
        coupling_matrix=cpl_matrix_2d,
        channel_matrix=ch_matrix,
        readout_directions=directions,
    )

    detector.signal.array = signal_2d



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Simple models to generate charge due to dark current process."""

import typing as t

import numpy as np

from pyxel.detectors import Detector
from pyxel.util import temporary_random_state


def calculate_dark_current(
    num_rows: int, num_cols: int, current: float, exposure_time: float
) -> np.ndarray:
    """Simulate dark current in a :term:`CCD`.

    Parameters
    ----------
    num_rows : int
        Number of rows for the generated image.
    num_cols : int
        Number of columns for the generated image.
    current : float
        Dark current, in electrons/pixel/second
    exposure_time : float
        Length of the simulated exposure, in seconds.

    Returns
    -------
    dark_im_array_2d: ndarray
        An array the same shape and dtype as the input containing dark counts
        in units of charge (e-).
    """
    # dark current for every pixel
    mean_dark_charge = current * exposure_time

    # This random number generation should change on each call.
    dark_im_array_2d = np.random.poisson(mean_dark_charge, size=(num_rows, num_cols))
    return dark_im_array_2d


@temporary_random_state
def dark_current(
    detector: Detector, dark_rate: float, seed: t.Optional[int] = None
) -> None:
    """Simulate dark current in a detector.

    Parameters
    ----------
    detector : Detector
        Any detector object.
    dark_rate : float
        Dark current, in electrons/pixel/second, which is the way
        manufacturers typically report it.
    seed: int, optional
    """

    exposure_time = detector.time_step
    geo = detector.geometry

    dark_current_array = calculate_dark_current(
        num_rows=geo.row,
        num_cols=geo.col,
        current=dark_rate,
        exposure_time=exposure_time,
    ).astype(
        float
    )  # type: np.ndarray

    detector.charge.add_charge_array(dark_current_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Simple Inter Pixel Capacitance model: https://iopscience.iop.org/article/10.1088/1538-3873/128/967/095001/pdf."""

import numpy as np
from astropy.convolution import convolve_fft

from pyxel.detectors import CMOS


def ipc_kernel(
    coupling: float, diagonal_coupling: float = 0.0, anisotropic_coupling: float = 0.0
) -> np.ndarray:
    """Return the IPC convolution kernel from the input coupling parameters.

    Parameters
    ----------
    coupling: float
    diagonal_coupling: float
    anisotropic_coupling: float

    Returns
    -------
    kernel: np.ndarray
    """

    if not diagonal_coupling < coupling:
        raise ValueError("Requirement diagonal_coupling <= coupling is not met.")
    if not anisotropic_coupling < coupling:
        raise ValueError("Requirement anisotropic_coupling <= coupling is not met.")
    if not 0 <= coupling + diagonal_coupling <= 0.25:
        raise ValueError("Requirement coupling + diagonal_coupling << 1 is not met.")

    kernel = np.array(
        [
            [diagonal_coupling, coupling - anisotropic_coupling, diagonal_coupling],
            [
                coupling + anisotropic_coupling,
                1 - 4 * (coupling + diagonal_coupling),
                coupling + anisotropic_coupling,
            ],
            [diagonal_coupling, coupling - anisotropic_coupling, diagonal_coupling],
        ],
        dtype=float,
    )

    return kernel


def compute_ipc_convolution(
    input: np.ndarray,
    coupling: float,
    diagonal_coupling: float,
    anisotropic_coupling: float,
) -> np.ndarray:
    """Compute convolution of the array with IPC kernel.

    Parameters
    ----------
    input: ndarray
    coupling: float
    diagonal_coupling: float
    anisotropic_coupling: float

    Returns
    -------
    array: ndarray
    """
    kernel = ipc_kernel(
        coupling=coupling,
        diagonal_coupling=diagonal_coupling,
        anisotropic_coupling=anisotropic_coupling,
    )

    # Convolution, extension on the edges with the mean value
    mean = np.mean(input)
    array = convolve_fft(input, kernel, boundary="fill", fill_value=mean)

    return array


def simple_ipc(
    detector: "CMOS",
    coupling: float,
    diagonal_coupling: float = 0.0,
    anisotropic_coupling: float = 0.0,
) -> None:
    """Convolve pixel array with the IPC kernel.

    Parameters
    ----------
    detector: CMOS
    coupling: float
    diagonal_coupling: float
    anisotropic_coupling: float

    Returns
    -------
    None
    """
    if not isinstance(detector, CMOS):
        raise TypeError("Expecting a CMOS object for detector.")

    array = compute_ipc_convolution(
        input=detector.pixel.array,
        coupling=coupling,
        diagonal_coupling=diagonal_coupling,
        anisotropic_coupling=anisotropic_coupling,
    )

    detector.pixel.array = array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Simple ADC model functions."""

import typing as t

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector


def apply_simple_adc(
    signal: np.ndarray, bit_resolution: int, voltage_range: t.Tuple[float, float]
) -> np.ndarray:
    """Apply digitization.

    Parameters
    ----------
    signal: np.ndarray
        Input signal.
    bit_resolution: int
        ADC bit resolution.
    voltage_range: tuple of floats
        ADC voltage range.

    Returns
    -------
    output: ndarray
    """
    bins = np.linspace(
        start=voltage_range[0], stop=voltage_range[1], num=2 ** bit_resolution - 1
    )
    output = np.digitize(x=signal, bins=bins, right=True)
    return output


def simple_adc(
    detector: Detector,
    data_type: Literal["uint16", "uint32", "uint64", "uint"] = "uint32",
) -> None:
    """Apply simple Analog to Digital conversion.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    data_type : str
        The desired data-type for the Image array. The data-type must be an unsigned integer.
        Valid values: 'uint16', 'uint32', 'uint64', 'uint'
        Invalid values: 'int16', 'int32', 'int64', 'int', 'float'...
    """

    bit_resolution = detector.characteristics.adc_bit_resolution
    voltage_range = detector.characteristics.adc_voltage_range

    try:
        d_type = np.dtype(data_type)  # type: np.dtype
    except TypeError as ex:
        raise TypeError(
            "Can not locate the type defined as `data_type` argument in yaml file."
        ) from ex

    if not issubclass(d_type.type, np.integer):
        raise TypeError("Expecting a signed/unsigned integer.")

    detector.image.array = np.asarray(
        apply_simple_adc(
            signal=detector.signal.array,
            bit_resolution=bit_resolution,
            voltage_range=voltage_range,
        ),
        dtype=d_type,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout electronics model."""

import numpy as np

from pyxel.detectors import Detector


def apply_amplify(
    signal_2d: np.ndarray,
    amplifier_gain: float,
) -> np.ndarray:
    """Apply gain from the output amplifier and signal processor.

    Parameters
    ----------
    signal_2d : ndarray
        2D signal to amplify. Unit: V
    amplifier_gain : float
        Gain of amplifier. Unit: V/V

    Returns
    -------
    ndarray
        2D amplified signal. Unit: V
    """
    amplified_signal_2d = signal_2d * amplifier_gain

    return amplified_signal_2d


def simple_amplifier(detector: Detector) -> None:
    """Amplify signal using gain from the output amplifier and the signal processor.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    """
    char = detector.characteristics

    amplified_signal_2d = apply_amplify(
        signal_2d=detector.signal.array,
        amplifier_gain=char.pre_amplification,
    )  # type: np.ndarray

    detector.signal.array = amplified_signal_2d



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel charge collection model."""

from pyxel.detectors import Detector


def simple_collection(detector: Detector) -> None:
    """Associate charge with the closest pixel.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    """
    detector.pixel.array += detector.charge.array



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Charge Distortion Model for :term:`CCDs<CCD>`.

============================
This is a function to run the upgraded :term:`CDM` :term:`CTI` model developed by Alex Short (ESA).

:requires: NumPy

:author: Alex Short
:author: Sami-Matias Niemi
:author: David Lucsanyi

Descriptions of model quantities in the paper:

Ne - number of electrons in a pixel
ne - electron density in the vicinity of the trap
Vc - volume of the charge cloud
Pr - the probability that the trap will release the electron into the sample
tau_c - capture time constant
Pc - capture probability (per vacant trap) as a function of the number of sample electrons Ne
NT - number of traps in the column,
NT = 2*nt*Vg*x  where x is the number of TDI transfers or the column length in pixels.
Nc - number of electrons captured by a given trap species during the transit of an integrating signal packet
N0 - initial trap occupancy
Nr - number of electrons released into the sample during a transit along the column
vg: assumed maximum geometrical volume electrons can occupy within a pixel (parallel)
svg: assumed maximum geometrical volume electrons can occupy within a pixel (serial)
t: constant TDI period (parallel)
st: constant TDI period (serial)
"""

import typing
from enum import Enum

import astropy.constants as const
import numba
import numpy as np
from typing_extensions import Literal

from pyxel.detectors import CCD


class CDMdirection(Enum):
    """:term:`CDM` direction class."""

    Parallel = "parallel"
    Serial = "serial"


def cdm(
    detector: CCD,
    direction: Literal["parallel", "serial"],
    beta: float,
    trap_release_times: typing.Sequence[float],
    trap_densities: typing.Sequence[float],
    sigma: typing.Sequence[float],
    full_well_capacity: typing.Optional[float] = None,
    max_electron_volume: float = 0.0,
    transfer_period: float = 0.0,
    charge_injection: bool = False,
    electron_effective_mass: float = 0.5,
) -> None:
    r"""Charge Distortion Model (CDM) model wrapper.

    Parameters
    ----------
    detector: CCD
        Pyxel CCD detector object.
    direction: literal
        Set ``"parallel"`` for :term:`CTI` in parallel direction or ``"serial"`` for :term:`CTI` in serial register.
    beta: float
        Electron cloud expansion coefficient :math:`\beta`.
    trap_release_times: sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    trap_densities: sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma: sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.
    full_well_capacity: float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    max_electron_volume: float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    transfer_period: float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    charge_injection: bool
        Enable charge injection (only used in ``"parallel"`` mode).
    electron_effective_mass: float
        Electron effective mass in the semiconductor lattice. Unit: 1 electron mass
    """

    e_effective_mass = electron_effective_mass * const.m_e.value

    if full_well_capacity is None:
        fwc_final = detector.characteristics.full_well_capacity
    else:
        fwc_final = full_well_capacity

    mode = CDMdirection(direction)

    if not isinstance(detector, CCD):
        # Later, this will be checked in when YAML configuration file is parsed
        raise TypeError("Expecting a `CCD` object for 'detector'.")

    if not (0.0 <= max_electron_volume <= 1.0):
        raise ValueError("'max_electron_volume' must be between 0.0 and 1.0.")
    if not (0.0 <= beta <= 1.0):
        raise ValueError("'beta' must be between 0.0 and 1.0.")
    if not (0.0 <= fwc_final <= 1.0e7):
        raise ValueError("'full_well_capacity' must be between 0 and 1e7.")
    if not (0.0 <= transfer_period <= 10.0):
        raise ValueError("'transfer_period' must be between 0.0 and 10.0.")
    if not (len(trap_densities) == len(trap_release_times) == len(sigma)):
        raise ValueError(
            "Length of 'sigma', 'trap_densities' and 'trap_release_times' not the same!"
        )
    if len(trap_release_times) == 0:
        raise ValueError("Expecting inputs for at least one trap species.")

    e_thermal_velocity = np.sqrt(
        3 * const.k_B.value * detector.environment.temperature / e_effective_mass
    )

    if mode == CDMdirection.Parallel:
        detector.pixel.array = run_cdm_parallel(
            array=detector.pixel.array,
            vg=max_electron_volume,
            t=transfer_period,
            fwc=fwc_final,
            vth=e_thermal_velocity,
            charge_injection=charge_injection,
            chg_inj_parallel_transfers=detector.geometry.row,
            beta=beta,
            tr=np.array(trap_release_times),
            nt=np.array(trap_densities),
            sigma=np.array(sigma),
        )

    elif mode == CDMdirection.Serial:
        detector.pixel.array = run_cdm_serial(
            array=detector.pixel.array,
            vg=max_electron_volume,
            t=transfer_period,
            fwc=fwc_final,
            vth=e_thermal_velocity,
            beta=beta,
            tr=np.array(trap_release_times),
            nt=np.array(trap_densities),
            sigma=np.array(sigma),
        )


@numba.njit(nogil=True)
def run_cdm_parallel(
    array: np.ndarray,
    beta: float,
    vg: float,
    t: float,
    fwc: float,
    vth: float,
    tr: np.ndarray,
    nt: np.ndarray,
    sigma: np.ndarray,
    charge_injection: bool = False,
    chg_inj_parallel_transfers: int = 0,
) -> np.ndarray:
    r"""Run :term:`CDM` in parallel direction.

    Parameters
    ----------
    array: ndarray
        Input array.
    beta: float
        Electron cloud expansion coefficient :math:`\beta`.
    vg: float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    t: float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    fwc: float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    vth: float
        Electron thermal velocity.
    tr: sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    nt: sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma: sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.
    charge_injection: bool
        Enable charge injection (only used in "parallel" mode).
    chg_inj_parallel_transfers: int
        Number of parallel transfers for charge injection.

    Returns
    -------
    array: ndarray
        Output array.
    """
    ydim, xdim = array.shape  # full signal array we want to apply cdm for
    kdim_p = len(nt)
    # np.clip(s, 0., fwc, s)      # full well capacity
    nt = nt / vg  # parallel trap density (traps / cm**3)
    # nt_p *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Parallel direction
    no = np.zeros((xdim, kdim_p))
    alpha_p = t * sigma * vth * fwc ** beta / (2.0 * vg)  # type: np.ndarray
    g_p = 2.0 * nt * vg / fwc ** beta  # type: np.ndarray
    for i in range(0, ydim):
        if charge_injection:
            gamma_p = (
                g_p * chg_inj_parallel_transfers
            )  # number of all transfers in parallel dir.
        else:
            gamma_p = g_p * i  # TODO: (i+1) ?????????????
        for k in range(kdim_p):
            for j in range(xdim):
                nc = 0.0
                if array[i, j] > 0.01:
                    nc = max(
                        (gamma_p[k] * array[i, j] ** beta - no[j, k])
                        / (gamma_p[k] * array[i, j] ** (beta - 1.0) + 1.0)
                        * (1.0 - np.exp(-1 * alpha_p[k] * array[i, j] ** (1.0 - beta))),
                        0.0,
                    )
                    no[j, k] += nc
                nr = no[j, k] * (1.0 - np.exp(-t / tr[k]))
                array[i, j] += -1 * nc + nr
                no[j, k] -= nr
                if array[i, j] < 0.01:
                    array[i, j] = 0.0

    return array


@numba.njit(nogil=True)
def run_cdm_serial(
    array: np.ndarray,
    beta: float,
    vg: float,
    t: float,
    fwc: float,
    vth: float,
    tr: np.ndarray,
    nt: np.ndarray,
    sigma: np.ndarray,
) -> np.ndarray:
    r"""Run :term:`CDM` in serial direction.

    Parameters
    ----------
    array: ndarray
        Input array.
    beta: float
        Electron cloud expansion coefficient :math:`\beta`.
    vg: float
        Maximum geometrical volume :math:`V_g` that electrons can occupy within a pixel. Unit: :math:`cm^3`.
    t: float
        Transfer period :math:`t` (TDI period). Unit: :math:`s`.
    fwc: float
        Full well capacity :math:`FWC`. Unit: :math:`e^-`.
    vth: float
        Electron thermal velocity.
    tr: sequence of float
        Trap release time constants :math:`\tau_r`. Unit: :math:`s`.
    nt: sequence of float
        Absolute trap densities :math:`n_t`. Unit: :math:`cm^{-3}`.
    sigma: sequence of float
        Trap capture cross section :math:`\sigma`. Unit: :math:`cm^2`.

    Returns
    -------
    array: ndarray
    """
    ydim, xdim = array.shape  # full signal array we want to apply cdm for
    kdim_s = len(nt)
    # np.clip(s, 0., fwc, s)      # full well capacity
    nt = nt / vg  # serial trap density (traps / cm**3)
    # nt_s *= rdose             # absolute trap density [per cm**3]

    # IMAGING (non-TDI) MODE
    # Serial direction

    sno = np.zeros((ydim, kdim_s))
    alpha_s = t * sigma * vth * fwc ** beta / (2.0 * vg)  # type: np.ndarray
    g_s = 2.0 * nt * vg / fwc ** beta
    for j in range(0, xdim):
        gamma_s = g_s * j  # TODO: (j+1) ?????????????
        for k in range(kdim_s):
            for i in range(ydim):
                nc = 0.0
                if array[i, j] > 0.01:
                    nc = max(
                        (gamma_s[k] * array[i, j] ** beta - sno[i, k])
                        / (gamma_s[k] * array[i, j] ** (beta - 1.0) + 1.0)
                        * (1.0 - np.exp(-1 * alpha_s[k] * array[i, j] ** (1.0 - beta))),
                        0.0,
                    )
                    sno[i, k] += nc
                nr = sno[i, k] * (1.0 - np.exp(-t / tr[k]))
                array[i, j] += -1 * nc + nr
                sno[i, k] -= nr
                if array[i, j] < 0.01:
                    array[i, j] = 0.0

    return array


#
# def plot_serial_profile(
#     data: np.ndarray, row: int, data2: t.Optional[np.ndarray] = None
# ) -> None:
#     """TBW.
#
#     :param data:
#     :param row:
#     :param data2:
#     """
#     ydim, xdim = data.shape
#     profile_x = list(range(ydim))
#     profile_y = data[row, :]
#     plt.title("Serial profile")
#     plt.plot(profile_x, profile_y, color="blue")
#     if data2 is not None:
#         profile_y_2 = data2[row, :]
#         plt.plot(profile_x, profile_y_2, color="red")
#
#
# def plot_parallel_profile(
#     data: np.ndarray, col: int, data2: t.Optional[np.ndarray] = None
# ) -> None:
#     """TBW.
#
#     :param data:
#     :param col:
#     :param data2:
#     """
#     ydim, xdim = data.shape
#     profile_x = list(range(xdim))
#     profile_y = data[:, col]
#     plt.title("Parallel profile")
#     plt.plot(profile_x, profile_y, color="blue")
#     if data2 is not None:
#         profile_y_2 = data2[:, col]
#         plt.plot(profile_x, profile_y_2, color="red")
#
#
# def plot_1d_profile(
#     array: np.ndarray, offset: int = 0, label: str = "", m: str = "-"
# ) -> None:
#     """Plot profile on log scale.
#
#     :param array:
#     :param offset:
#     :param label:
#     :param m:
#     """
#     x = list(range(offset, offset + len(array)))
#     # plt.title('Parallel profile, charge injection')
#     plt.semilogy(x, array, m, label=label)
#     if label:
#         plt.legend()
#
#
# def plot_1d_profile_lin(
#     array: np.ndarray,
#     offset: int = 0,
#     label: str = "",
#     m: str = "-",
#     col: t.Optional[str] = None,
# ) -> None:
#     """TBW.
#
#     :param array:
#     :param offset:
#     :param label:
#     :param m:
#     :param col:
#     """
#     x = list(range(offset, offset + len(array)))
#     # plt.title('Parallel profile, charge injection')
#     plt.plot(x, array, m, label=label, color=col)
#     if label:
#         plt.legend()
#
#
# def plot_1d_profile_with_err(
#     array: np.ndarray, error: np.ndarray, offset: int = 0, label: str = ""
# ) -> None:
#     """TBW.
#
#     :param array:
#     :param error:
#     :param offset:
#     :param label:
#     """
#     x = list(range(offset, offset + len(array)))
#     plt.title("Parallel profile with error, charge injection")
#     plt.errorbar(x, array, error, label=label, linestyle="None", marker=".")
#     if label:
#         plt.legend()
#
#
# def plot_residuals(
#     data: np.ndarray, data2: np.ndarray, label: str = ""
# ) -> None:  # col='magenta',
#     """TBW.
#
#     :param data:
#     :param data2:
#     # :param col:
#     :param label:
#     """
#     x = list(range(len(data)))
#     # plt.title('Residuals of fitted and target parallel CTI profiles')
#     residuals = np.around(data - data2, decimals=5)
#     # residuals = data-data2
#     # plt.plot(x, residuals, '.', color=col, label=label)
#     plt.plot(x, residuals, ".", label=label)
#     plt.legend()
#
#
# def plot_image(data: np.ndarray) -> None:
#     """TBW.
#
#     :param data:
#     """
#     plt.imshow(data, cmap=plt.gray())  # , interpolation='nearest')
#     plt.xlabel("x - serial direction")
#     plt.ylabel("y - parallel direction")
#     plt.title("CCD image with CTI")
#     plt.colorbar()


# # @numba.jit(nopython=True, nogil=True, parallel=True)
# def optimized_cdm(
#     s: np.ndarray,
#     beta_p: float,
#     beta_s: float,
#     vg: float,
#     svg: float,
#     t: float,
#     st: float,
#     fwc: float,
#     sfwc: float,
#     vth: float,
#     tr_p: np.ndarray,
#     tr_s: np.ndarray,
#     nt_p: np.ndarray,
#     nt_s: np.ndarray,
#     sigma_p: np.ndarray,
#     sigma_s: np.ndarray,
#     charge_injection: bool = False,
#     chg_inj_parallel_transfers: int = 0,
#     parallel_cti: bool = True,
#     serial_cti: bool = True,
# ) -> np.ndarray:
#     """CDM model.
#
#     Done by Patricia Liebing. Not yet test or compared to the original 'run_cdm' function.
#
#     :param s: np.ndarray
#     :param dob:
#     :param beta_p: electron cloud expansion coefficient (parallel)
#     :param beta_s: electron cloud expansion coefficient (serial)
#     :param vg: assumed maximum geometrical volume electrons can occupy within
#                a pixel (parallel)
#     :param svg: assumed maximum geometrical volume electrons can occupy within
#                 a pixel (serial)
#     :param t: constant TDI period (parallel)
#     :param st: constant TDI period (serial)
#     :param fwc:
#     :param sfwc:
#     :param vth:
#     :param charge_injection:
#     :param chg_inj_parallel_transfers:
#     :param sigma_p:
#     :param sigma_s:
#     :param tr_p:
#     :param tr_s:
#     :param nt_p: number of traps per electron cloud (and not pixel!)
#                  in parallel direction
#     :param nt_s: number of traps per electron cloud (and not pixel!)
#                  in serial direction
#     :param parallel_cti:
#     :param serial_cti:
#     :return:
#     """
#     ydim, xdim = s.shape  # full signal array we want to apply cdm for
#
#     kdim_p = len(nt_p)
#     kdim_s = len(nt_s)
#
#     # np.clip(s, 0., fwc, s)      # full well capacity
#
#     nt_p = nt_p / vg  # parallel trap density (traps / cm**3)
#     nt_s = nt_s / svg  # serial trap density (traps / cm**3)
#
#     # nt_p *= rdose             # absolute trap density [per cm**3]
#     # nt_s *= rdose             # absolute trap density [per cm**3]
#
#     # IMAGING (non-TDI) MODE
#     # Parallel direction
#     if parallel_cti:
#         no = np.zeros((xdim, kdim_p))
#         alpha_p = t * sigma_p * vth * fwc ** beta_p / (2.0 * vg)  # type: np.ndarray
#         g_p = 2.0 * nt_p * vg / fwc ** beta_p  # type: np.ndarray
#         for i in range(0, ydim):
#             if charge_injection:
#                 gamma_p = (
#                     g_p * chg_inj_parallel_transfers
#                 )  # number of all transfers in parallel dir.
#             else:
#                 gamma_p = g_p * (i + 1)
#             for k in range(kdim_p):
#                 nc = np.zeros(xdim)
#                 nctest = np.zeros(xdim)
#                 sclm = s[i, :]
#                 idx = np.where(sclm > 0.01)
#                 nok = no[:, k]
#                 nc[idx] += (
#                     (gamma_p[k] * sclm[idx] ** beta_p - nok[idx])
#                     / (gamma_p[k] * sclm[idx] ** (beta_p - 1.0) + 1.0)
#                     * (1.0 - np.exp(-1 * alpha_p[k] * sclm[idx] ** (1.0 - beta_p)))
#                 )
#                 no[:, k] += np.fmax(nctest, nc)
#                 nr = no[:, k] * (1.0 - np.exp(-t / tr_p[k]))
#                 s[i, :] += -1 * nc + nr
#                 no[:, k] -= nr
#
#     # IMAGING (non-TDI) MODE
#     # Serial direction
#     if serial_cti:
#         sno = np.zeros((ydim, kdim_s))
#         alpha_s = st * sigma_s * vth * sfwc ** beta_s / (2.0 * svg)  # type: np.ndarray
#         g_s = 2.0 * nt_s * svg / sfwc ** beta_s  # type: np.ndarray
#         for j in range(0, xdim):
#             gamma_s = g_s * (j + 1)
#             for k in range(kdim_s):
#                 ncs = np.zeros(ydim)
#                 ncstest = np.zeros(ydim)
#                 srow = s[:, j]
#                 idxs = np.where(srow > 0.01)
#                 snok = sno[:, k]
#                 ncs[idxs] += (
#                     (gamma_s[k] * srow[idxs] ** beta_s - snok[idxs])
#                     / (gamma_s[k] * srow[idxs] ** (beta_s - 1.0) + 1.0)
#                     * (1.0 - np.exp(-1 * alpha_s[k] * srow[idxs] ** (1.0 - beta_s)))
#                 )
#                 sno[:, k] += np.fmax(ncstest, ncs)
#                 nrs = sno[:, k] * (1.0 - np.exp(-st / tr_s[k]))
#                 s[:, j] += -1 * ncs + nrs
#                 sno[:, k] -= nrs
#
#     return s



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
from pyxel.detectors import CMOSCharacteristics


class MKIDCharacteristics(CMOSCharacteristics):
    """Characteristic attributes of a :term:`MKID`-based detector.

    Parameters
    ----------
    quantum_efficiency: float
        Quantum efficiency.
    charge_to_volt_conversion: float
        Sensitivity of charge readout. Unit: V/e-
    pre_amplification: float
        Gain of pre-amplifier. Unit: V/V
    adc_gain: float
        Gain of the Analog-Digital Converter. Unit: ADU/V
    full_well_capacity: float
        Full well capacity. Unit: e-
    """



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Utility functions."""

import logging
import typing as t

import numpy as np

# flake8: noqa
# from pyxel.util.outputs import image, numpy_array, hist_plot, graph_plot, show_plots

from pyxel.util.memory import get_size, memory_usage_details
from pyxel.util.examples import download_examples
from pyxel.util.timing import time_pipeline
from pyxel.util.add_model import create_model
from pyxel.util.random import temporary_random_state
from pyxel.util.image import fit_into_array, load_cropped_and_aligned_image

__all__ = [
    "convert_to_int",
    "round_convert_to_int",
    "PipelineAborted",
    "LogFilter",
]


class PipelineAborted(Exception):
    """Exception to force the pipeline to stop processing."""

    def __init__(
        self,
        message: t.Optional[str] = None,
        # errors=None
    ):
        super().__init__(message)
        # self.errors = errors


def round_convert_to_int(input_array: np.ndarray) -> np.ndarray:
    """Round list of floats in numpy array and convert to integers.

    Use on data before adding into DataFrame.

    :param input_array: numpy array object OR numpy array (float, int)
    :return:
    """
    array = input_array.astype(float)
    array = np.rint(array)
    array = array.astype(int)
    return array


def convert_to_int(input_array: np.ndarray) -> np.ndarray:
    """Convert numpy array to integer.

    Use on data after getting it from DataFrame.

    :param input_array: numpy array object OR numpy array (float, int)
    :return:
    """
    return input_array.astype(int)


class LogFilter(logging.Filter):
    """TBW."""

    def filter(self, record):
        """Filter or modify record."""
        if record.threadName == "MainThread" or record.threadName.endswith("-0"):
            return True
        else:
            return False



# import pyxel
from pyxel.detectors import Detector

# @pyxel.register(group='charge_generation', name='my_class_model')
# class MyClassModel:
#     def __call__(self, detector: Detector, level: int, noise: float = 2.0) -> Detector:
#         setattr(detector, "level", level)
#         setattr(detector, "noise", noise)
#         return detector


# @pyxel.register(group='charge_generation', name='my_other_class_model')
# class MyOtherClassModel:
#     def __call__(self, detector: Detector, std: float = 2.0) -> Detector:
#         setattr(detector, "std", std)
#         return detector


# def my_function_model(detector: Detector, level, noise: float = 2.0) -> Detector:
#     # set a new attribute so it can be checked later
#     setattr(detector, "level", level)
#     setattr(detector, "noise", noise)
#     return detector


# pyxel.register(group='charge_generation', maybe_func=my_function_model)


# @pyxel.register(group='charge_generation', name='my_dec_model_class')
class MyDecoratedModel:
    def __call__(self, detector: Detector, class_std=1.0):
        setattr(detector, "class_std", class_std)
        return detector


# @pyxel.register(group='charge_generation', name='my_dec_model_func')
# def my_decorated_function(detector: Detector, func_std=2.0):
#     setattr(detector, "func_std", func_std)
#     return detector


registry_map = {
    "photon_generation": [
        {"func": "pyxel.models.photon_generation.load_image"},
        {"func": "pyxel.models.photon_generation.illumination"},
        {"func": "pyxel.models.photon_generation.shot_noise"},
    ],
    "optics": [],
    "charge_generation": [
        {
            "func": "pyxel.models.charge_generation.simple_conversion",
            "name": "photoelectrons",
        },
        {"func": "pyxel.models.run_tars", "name": "tars"},
    ],
    "charge_collection": [
        {
            "func": "pyxel.models.charge_collection.fix_pattern_noise",
            "type": "ccd",
        },
        {
            "func": "pyxel.models.charge_collection.simple_full_well",
            "name": "full_well",
        },
    ],
    "charge_transfer": [{"func": "pyxel.models.charge_transfer.cdm", "type": "ccd"}],
    "charge_measurement": [
        {
            "func": "pyxel.models.charge_measurement.output_node_noise",
            "type": "ccd",
        },
        {
            "func": "pyxel.models.signal_transfer.nghxrg.nghxrg.ktc_bias_noise",
            "type": "cmos",
        },
        {"func": "pyxel.models.signal_transfer.nghxrg.nghxrg.white_read_noise"},
    ],
    "signal_transfer": [
        {
            "func": "pyxel.models.signal_transfer.nghxrg.nghxrg.corr_pink_noise",
            "type": "cmos",
        },
        {
            "func": "pyxel.models.signal_transfer.nghxrg.nghxrg.uncorr_pink_noise",
            "type": "cmos",
        },
        {
            "func": "pyxel.models.signal_transfer.nghxrg.nghxrg.acn_noise",
            "type": "cmos",
        },
    ],
    "readout_electronics": [
        {
            "func": "pyxel.models.signal_transfer.nghxrg.nghxrg.pca_zero_noise",
            "type": "cmos",
        }
    ],
}



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Linearity models."""
import typing as t

import numpy as np

from pyxel.detectors import Detector


def compute_poly_linearity(
    array_2d: np.ndarray,
    coefficients: t.Sequence[float],
) -> np.ndarray:
    """Add non-linearity to an array of values following a polynomial function.

    Parameters
    ----------
    array_2d: ndarray
        Input array.
    coefficients: list of float
        Coefficients of the polynomial function.

    Returns
    -------
    signal: np.ndarray
    """

    polynomial_function = np.polynomial.polynomial.Polynomial(coefficients)

    non_linear_signal = polynomial_function(array_2d)

    return non_linear_signal


def output_node_linearity_poly(
    detector: Detector,
    coefficients: t.Sequence[float],
) -> None:
    """Add non-linearity to signal array to simulate the non-linearity of the output node circuit.

    The non-linearity is simulated by a polynomial function. The user specifies the polynomial coefficients.

    detector Signal unit: Volt

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    coefficients: list of float
        Coefficient of the polynomial function.
    """
    if len(coefficients) == 0:
        raise ValueError("Length of coefficient list should be more than 0.")

    signal_mean_array = detector.signal.array.astype("float64")
    signal_non_linear = compute_poly_linearity(
        array_2d=signal_mean_array, coefficients=coefficients
    )

    if np.any(signal_non_linear < 0):
        raise ValueError(
            "Signal array contains negative values after applying non-linearity model!"
        )

    detector.signal.array = signal_non_linear



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
from pyxel.detectors import Geometry


class CCDGeometry(Geometry):
    """Geometrical attributes of a :term:`CCD` detector.

    Parameters
    ----------
    row: int
        Number of pixel rows.
    col: int
        Number of pixel columns.
    total_thickness: float
        Thickness of detector. Unit: um
    pixel_vert_size: float
        Vertical dimension of pixel. Unit: um
    pixel_horz_size: float
        Horizontal dimension of pixel. Unit: um
    """



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Functions to add new models."""

import logging
import os
import shutil
import sys
import time
import typing as t


def create_model(newmodel: str) -> None:
    """Create a new module using pyxel/templates/MODELTEMPLATE.py.

    Parameters
    ----------
    newmodel: modeltype/modelname

    Returns
    -------
    None
    """

    location, model_name = get_name_and_location(newmodel)

    # Is not working on UNIX AND Windows if I do not use os.path.abspath
    path = os.path.abspath(os.getcwd() + "/pyxel/models/" + location + "/")
    template_string = "_TEMPLATE"
    template_location = "_LOCATION"

    # Copying the template with the user defined model_name instead
    import pyxel

    src = os.path.abspath(os.path.dirname(pyxel.__file__) + "/templates/")
    dest = os.path.abspath(
        os.path.dirname(pyxel.__file__) + "/models/" + location + "/"
    )

    try:
        os.mkdir(dest)
        # Replacing all of template in filenames and directories by model_name
        for dirpath, subdirs, files in os.walk(src):
            for x in files:
                pathtofile = os.path.join(dirpath, x)
                new_pathtofile = os.path.join(
                    dest, x.replace(template_string, model_name)
                )
                shutil.copy(pathtofile, new_pathtofile)
                # Open file in the created copy
                with open(new_pathtofile, "r") as file_tochange:
                    # Replace any mention of template by model_name
                    new_contents = file_tochange.read().replace(
                        template_string, model_name
                    )
                    new_contents = new_contents.replace(template_location, location)
                    new_contents = new_contents.replace("%(date)", time.ctime())
                with open(new_pathtofile, "w+") as file_tochange:
                    file_tochange.write(new_contents)
                # Close the file other we can't rename it
                file_tochange.close()

            for x in subdirs:
                pathtofile = os.path.join(dirpath, x)
                os.mkdir(pathtofile.replace(template_string, model_name))
            logging.info("Module " + model_name + " created.")
        print("Module " + model_name + " created in " + path + ".")
    except FileExistsError:
        logging.info(f"{dest} already exists, folder not created")
    # Directories are the same
    except shutil.Error as e:
        logging.critical("Error while duplicating " + template_string + ": %s" % e)
    # Any error saying that the directory doesn't exist
    except OSError as e:
        logging.critical(model_name + " not created. Error: %s" % e)
    return None


def get_name_and_location(newmodel: str) -> t.Tuple[str, str]:
    """Get name and location of new model from string modeltype/modelname.

    Parameters
    ----------
    newmodel: str

    Returns
    -------
    location: str
    model_name: str
    """

    try:
        arguments = newmodel.split("/")
        location = f"{arguments[0]}"
        model_name = f"{arguments[1]}"
    except Exception:
        sys.exit(
            f"""
        Can't create model {arguments}, please use location/newmodelname
        as an argument for creating a model
        """
        )
    return location, model_name



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""CLI to run Pyxel."""
import logging
import sys
import time
import typing as t
from pathlib import Path

import click
import dask
import pandas as pd
import xarray as xr
from matplotlib import pyplot as plt

from pyxel import Configuration
from pyxel import __version__ as version
from pyxel import load, outputs, save
from pyxel.calibration import Calibration, CalibrationResult
from pyxel.detectors import CCD, CMOS, MKID, Detector
from pyxel.exposure import Exposure
from pyxel.observation import Observation, ObservationResult
from pyxel.pipelines import DetectionPipeline, Processor
from pyxel.util import create_model, download_examples

if t.TYPE_CHECKING:
    from pyxel.outputs import CalibrationOutputs, ExposureOutputs, ObservationOutputs


def exposure_mode(
    exposure: "Exposure",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> xr.Dataset:
    """Run an 'exposure' pipeline.

    For more information, see :ref:`exposure_mode`.

    Parameters
    ----------
    exposure: Exposure
    detector: Detector
    pipeline: DetectionPipeline

    Returns
    -------
    Dataset
        An multi-dimensional array database from `xarray <https://xarray.pydata.org>`_.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run an exposure pipeline

    >>> dataset = pyxel.exposure_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> dataset
    <xarray.Dataset>
    Dimensions:       (readout_time: 1, y: 450, x: 450)
    Coordinates:
      * readout_time  (readout_time) int64 1
      * y             (y) int64 0 1 2 3 4 5 6 7 ... 442 443 444 445 446 447 448 449
      * x             (x) int64 0 1 2 3 4 5 6 7 ... 442 443 444 445 446 447 448 449
    Data variables:
        image         (readout_time, y, x) uint16 9475 9089 8912 ... 9226 9584 10079
        signal        (readout_time, y, x) float64 3.159 3.03 2.971 ... 3.195 3.36
        pixel         (readout_time, y, x) float64 1.053e+03 1.01e+03 ... 1.12e+03
    """

    logging.info("Mode: Exposure")

    exposure_outputs = exposure.outputs  # type: ExposureOutputs

    detector.set_output_dir(exposure_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    result = exposure.run_exposure(processor=processor)

    if exposure_outputs.save_exposure_data:
        exposure_outputs.save_exposure_outputs(dataset=result)

    return result


def observation_mode(
    observation: "Observation",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> "ObservationResult":
    """Run an 'observation' pipeline.

    For more information, see :ref:`observation_mode`.

    Parameters
    ----------
    observation: Observation
    detector: Detector
    pipeline: DetectionPipeline

    Returns
    -------
    ObservationResult
        Result.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run an observation pipeline

    >>> result = pyxel.observation_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> result
    ObservationResult(...)
    """
    logging.info("Mode: Observation")

    observation_outputs = observation.outputs  # type: ObservationOutputs
    detector.set_output_dir(observation_outputs.output_dir)  # TODO: Remove this

    # TODO: This should be done during initializing of object `Configuration`
    # parametric_outputs.params_func(parametric)

    processor = Processor(detector=detector, pipeline=pipeline)

    result = observation.run_observation(processor=processor)

    if observation_outputs.save_observation_data:
        observation_outputs.save_observation_datasets(
            result=result, mode=observation.parameter_mode
        )

    return result


def calibration_mode(
    calibration: "Calibration",
    detector: Detector,
    pipeline: "DetectionPipeline",
    compute_and_save: bool = True,
) -> t.Tuple[xr.Dataset, pd.DataFrame, pd.DataFrame, t.Sequence]:
    """Run a 'calibration' pipeline.

    For more information, see :ref:`calibration_mode`.

    Parameters
    ----------
    calibration: Calibration
    detector: Detector
    pipeline: DetectionPipeline
    compute_and_save: bool

    Returns
    -------
    tuple of Dataset, DataFrame, DataFrame, Sequence

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run a calibration pipeline

    >>> ds, processors, logs, filenames = pyxel.calibration_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )

    >>> ds
    <xarray.Dataset>
    Dimensions:              (island: 2, evolution: 2, param_id: 4, id_processor: 1, readout_time: 1, y: 100, x: 100)
    Coordinates:
      * island               (island) int64 0 1
      * evolution            (evolution) int64 0 1
      * id_processor         (id_processor) int64 0
      * readout_time         (readout_time) int64 1
      * y                    (y) int64 0 1 2 3 4 5 6 7 8 ... 92 93 94 95 96 97 98 99
      * x                    (x) int64 0 1 2 3 4 5 6 7 8 ... 92 93 94 95 96 97 98 99
      * param_id             (param_id) int64 0 1 2 3
    Data variables:
        champion_fitness     (evolution, island) float64 4.759e+06 ... 4.533e+06
        champion_decision    (evolution, island, param_id) float64 0.08016 ... 7....
        champion_parameters  (evolution, island, param_id) float64 0.08016 ... 7....
        simulated_image      (island, id_processor, readout_time, y, x) uint16 48...
        simulated_signal     (island, id_processor, readout_time, y, x) float64 4...
        simulated_pixel      (island, id_processor, readout_time, y, x) float64 6...
        target               (id_processor, y, x) >f8 4.834e+03 ... 4.865e+03
    Attributes:
        num_islands:      2
        population_size:  20
        num_evolutions:   2
        generations:      5
        topology:         unconnected
        result_type:      ResultType.Image

    >>> processors
           island  id_processor                                          processor
    0       0             0  Delayed('apply_parameters-c5da1649-766f-4ecb-a...
    1       1             0  Delayed('apply_parameters-c16f998f-f52f-4beb-b...

    >>> logs
        num_generations  ...  global_num_generations
    0                 1  ...                       1
    1                 2  ...                       2
    2                 3  ...                       3
    3                 4  ...                       4
    4                 5  ...                       5
    ..              ...  ...                     ...
    15                1  ...                       6
    16                2  ...                       7
    17                3  ...                       8
    18                4  ...                       9
    19                5  ...                      10

    >>> filenames
    []
    """
    logging.info("Mode: Calibration")

    calibration_outputs = calibration.outputs  # type: CalibrationOutputs
    detector.set_output_dir(calibration_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    ds_results, df_processors, df_all_logs = calibration.run_calibration(
        processor=processor, output_dir=calibration_outputs.output_dir
    )

    # TODO: Save the processors from 'df_processors'
    # TODO: Generate plots from 'ds_results'

    # TODO: Do something with 'df_all_logs' ?

    # TODO: create 'output' object with .calibration_outputs
    # TODO: use 'fitting.get_simulated_data' ==> np.ndarray

    # geometry = processor.detector.geometry
    # calibration.post_processing(
    #     champions=champions,
    #     output=calibration_outputs,
    #     row=geometry.row,
    #     col=geometry.col,
    # )
    filenames = calibration.post_processing(
        ds=ds_results, df_processors=df_processors, output=calibration_outputs
    )

    if compute_and_save:

        computed_ds, df_processors, df_logs, filenames = dask.compute(
            ds_results, df_processors, df_all_logs, filenames
        )

        if calibration_outputs.save_calibration_data:
            calibration_outputs.save_calibration_outputs(
                dataset=computed_ds, logs=df_logs
            )
            print(f"Saved calibration outputs to {calibration_outputs.output_dir}")

        result = CalibrationResult(
            dataset=computed_ds,
            processors=df_processors,
            logs=df_logs,
            filenames=filenames,
        )

    else:
        result = CalibrationResult(
            dataset=ds_results,
            processors=df_processors,
            logs=df_all_logs,
            filenames=filenames,
        )

    return result


def output_directory(configuration: Configuration) -> Path:
    """Return the output directory from the configuration.

    Parameters
    ----------
    configuration

    Returns
    -------
    output_dir
    """
    if isinstance(configuration.exposure, Exposure):
        output_dir = configuration.exposure.outputs.output_dir
    elif isinstance(configuration.calibration, Calibration):
        output_dir = configuration.calibration.outputs.output_dir
    elif isinstance(configuration.observation, Observation):
        output_dir = configuration.observation.outputs.output_dir
    else:
        raise (TypeError("Outputs not initialized."))
    return output_dir


def run(
    input_filename: t.Union[str, Path], random_seed: t.Optional[int] = None
) -> None:
    """Run a YAML configuration file.

    For more information, see :ref:`running_modes`.

    Parameters
    ----------
    input_filename : str or Path
    random_seed : int, optional

    Examples
    --------
    >>> import pyxel
    >>> pyxel.run("configuration.yaml")
    """
    logging.info("Pyxel version %s", version)
    logging.info("Pipeline started.")

    start_time = time.time()

    configuration = load(
        Path(input_filename).expanduser().resolve()
    )  # type: Configuration

    output_dir = output_directory(configuration)

    save(input_filename=input_filename, output_dir=output_dir)

    pipeline = configuration.pipeline  # type: DetectionPipeline

    if isinstance(configuration.ccd_detector, CCD):
        detector = configuration.ccd_detector  # type: t.Union[CCD, CMOS, MKID]
    elif isinstance(configuration.cmos_detector, CMOS):
        detector = configuration.cmos_detector
    elif isinstance(configuration.mkid_detector, MKID):
        detector = configuration.mkid_detector
    else:
        raise NotImplementedError("Detector is not defined in YAML config. file!")

    if isinstance(configuration.exposure, Exposure):
        exposure = configuration.exposure  # type: Exposure
        exposure_mode(
            exposure=exposure,
            detector=detector,
            pipeline=pipeline,
        )

    elif isinstance(configuration.calibration, Calibration):

        calibration = configuration.calibration  # type: Calibration
        _ = calibration_mode(
            calibration=calibration,
            detector=detector,
            pipeline=pipeline,
        )

    elif isinstance(configuration.observation, Observation):
        observation = configuration.observation  # type: Observation
        observation_mode(
            observation=observation,
            detector=detector,
            pipeline=pipeline,
        )

    else:
        raise NotImplementedError("Please provide a valid simulation mode !")

    logging.info("Pipeline completed.")
    logging.info("Running time: %.3f seconds" % (time.time() - start_time))
    # Closing the logger in order to be able to move the file in the output dir
    logging.shutdown()
    outputs.save_log_file(output_dir)
    plt.close()


# TODO: Add an option to display colors ?
@click.group()
@click.version_option(version=version)
def main():
    """Pyxel detector simulation framework.

    Pyxel is a detector simulation framework, that can simulate a variety of
    detector effects (e.g., cosmic rays, radiation-induced :term:`CTI` in :term:`CCDs<CCD>`, persistence
    in :term:`MCT`, charge diffusion, crosshatches, noises, crosstalk etc.) on a given image.
    """


@main.command(name="download-examples")
@click.argument("folder", type=click.Path(), default="pyxel-examples", required=False)
@click.option("-f", "--force", is_flag=True, help="Force flag for saving the examples.")
def download_pyxel_examples(folder, force: bool):
    """Install examples to a specified directory.

    Default folder is './pyxel-examples'.
    """
    download_examples(foldername=folder, force=force)


@main.command(name="create-model")
@click.argument("model_name", type=str)
def create_new_model(model_name: str):
    """Create a new model.

    Use: arg1/arg2. Create a new module in ``pyxel/models/arg1/arg2`` using a template
    (``pyxel/templates/MODELTEMPLATE.py``)
    """
    create_model(newmodel=model_name)


@main.command(name="run")
@click.argument("config", type=click.Path(exists=True))
@click.option(
    "-v",
    "--verbosity",
    count=True,
    show_default=True,
    help="Increase output verbosity (-v/-vv/-vvv)",
)
def run_config(config: str, verbosity: int):
    """Run Pyxel with a ``YAML`` configuration file."""
    logging_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][
        min(verbosity, 3)
    ]
    log_format = (
        "%(asctime)s - %(name)s - %(threadName)30s - %(funcName)30s \t %(message)s"
    )
    logging.basicConfig(
        filename="pyxel.log",
        level=logging_level,
        format=log_format,
        datefmt="%d-%m-%Y %H:%M:%S",
    )

    # If user wants the log in stdout AND in file, use the three lines below
    stream_stdout = logging.StreamHandler(sys.stdout)
    stream_stdout.setFormatter(logging.Formatter(log_format))
    logging.getLogger().addHandler(stream_stdout)

    run(input_filename=config)


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

import operator
import typing as t
from pathlib import Path

from typing_extensions import Literal

from pyxel.observation import ParameterMode
from pyxel.outputs import Outputs

if t.TYPE_CHECKING:
    from pyxel.observation import ObservationResult

    class SaveToFile(t.Protocol):
        """TBW."""

        def __call__(
            self, data: t.Any, name: str, with_auto_suffix: bool = True
        ) -> Path:
            """TBW."""
            ...


ValidName = Literal[
    "detector.image.array", "detector.signal.array", "detector.pixel.array"
]
ValidFormat = Literal["fits", "hdf", "npy", "txt", "csv", "png"]


class ObservationOutputs(Outputs):
    """TBW."""

    def __init__(
        self,
        output_folder: t.Union[str, Path],
        save_data_to_file: t.Optional[
            t.Sequence[t.Mapping[ValidName, t.Sequence[ValidFormat]]]
        ] = None,
        save_observation_data: t.Optional[
            t.Sequence[t.Mapping[str, t.Sequence[str]]]
        ] = None,
    ):
        super().__init__(
            output_folder=output_folder, save_data_to_file=save_data_to_file
        )

        self.save_observation_data = (
            save_observation_data
        )  # type: t.Optional[t.Sequence[t.Mapping[str, t.Sequence[str]]]]

    def save_observation_datasets(
        self, result: "ObservationResult", mode: "ParameterMode"
    ) -> None:
        """Save the result datasets from parametric mode on disk.

        Parameters
        ----------
        result: Result
        mode: ParameterMode

        Returns
        -------
        None
        """

        dataset_names = ("dataset", "parameters", "logs")

        save_methods = {"nc": self.save_to_netcdf}  # type: t.Dict[str, SaveToFile]

        if self.save_observation_data is not None:

            for (
                dct
            ) in self.save_observation_data:  # type: t.Mapping[str, t.Sequence[str]]
                first_item, *_ = dct.items()
                obj, format_list = first_item

                if obj not in dataset_names:
                    raise ValueError(
                        "Please specify a valid result dataset names ('dataset', 'parameters', 'logs')."
                    )

                if mode == ParameterMode.Sequential and obj == "dataset":
                    dct = operator.attrgetter(obj)(result)
                    for key, value in dct.items():

                        if format_list is not None:
                            for out_format in format_list:

                                if out_format not in save_methods.keys():
                                    raise ValueError(
                                        "Format "
                                        + out_format
                                        + " not a valid save method!"
                                    )

                                func = save_methods[out_format]
                                func(data=value, name=obj + "_" + key)

                else:
                    ds = operator.attrgetter(obj)(result)

                    if format_list is not None:
                        for out_format in format_list:

                            if out_format not in save_methods.keys():
                                raise ValueError(
                                    "Format " + out_format + " not a valid save method!"
                                )

                            func = save_methods[out_format]
                            func(data=ds, name=obj)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""
import typing as t

from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.util import load_cropped_and_aligned_image


def load_image(
    detector: Detector,
    image_file: str,
    position: t.Tuple[int, int] = (0, 0),
    align: t.Optional[
        Literal["center", "top_left", "top_right", "bottom_left", "bottom_right"]
    ] = None,
    convert_to_photons: bool = False,
    multiplier: float = 1.0,
    time_scale: float = 1.0,
    bit_resolution: t.Optional[int] = None,
) -> None:
    r"""Load FITS file as a numpy array and add to the detector as input image.

    Parameters
    ----------
    detector: Detector
    image_file: str
        Path to image file.
    position: tuple
        Indices of starting row and column, used when fitting image to detector.
    align: Literal
        Keyword to align the image to detector. Can be any from:
        ("center", "top_left", "top_right", "bottom_left", "bottom_right")
    convert_to_photons: bool
        If ``True``, the model converts the values of loaded image array from ADU to
        photon numbers for each pixel using the Photon Transfer Function:
        :math:`PTF = quantum\_efficiency \cdot charge\_to\_voltage\_conversion
        \cdot pre\_amplification \cdot adc\_factor`.
    multiplier: float
        Multiply photon array level with a custom number.
    time_scale: float
        Time scale of the photon flux, default is 1 second. 0.001 would be ms.
    bit_resolution: int
        Bit resolution of the loaded image.
    """

    shape = (detector.geometry.row, detector.geometry.col)
    position_y, position_x = position

    image = load_cropped_and_aligned_image(
        filename=image_file,
        shape=shape,
        align=align,
        position_x=position_x,
        position_y=position_y,
    )

    detector.input_image = image
    photon_array = image

    if convert_to_photons:

        if not bit_resolution:
            raise ValueError(
                "Bit resolution of the input image has to be specified for converting to photons."
            )

        cht = detector.characteristics
        adc_multiplier = cht.adc_voltage_range[1] / 2 ** bit_resolution

        photon_array = (
            photon_array
            * adc_multiplier
            / (
                cht.quantum_efficiency
                * cht.charge_to_volt_conversion
                * cht.pre_amplification
            )
        )
    photon_array = photon_array * (detector.time_step / time_scale) * multiplier

    try:
        detector.photon.array += photon_array
    except ValueError as ex:
        raise ValueError("Shapes of arrays do not match") from ex



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

import typing as t

from pyxel.util.memory import get_size


class Environment:
    """Environmental attributes of the detector.

    Parameters
    ----------
    temperature: float, optional
        Temperature of the detector. Unit: K
    """

    def __init__(self, temperature: t.Optional[float] = None):
        if temperature and not (0.0 <= temperature <= 1000.0):
            raise ValueError("'temperature' must be between 0.0 and 1000.0.")

        self._temperature = temperature

        self._numbytes = 0

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}(temperature={self._temperature!r})"

    @property
    def temperature(self) -> float:
        """Get Temperature of the detector."""
        if self._temperature:
            return self._temperature
        else:
            raise ValueError("'temperature' not specified in detector environment.")

    @temperature.setter
    def temperature(self, value: float) -> None:
        """Set Temperature of the detector."""
        if not (0.0 <= value <= 1000.0):
            raise ValueError("'temperature' must be between 0.0 and 1000.0.")

        self._temperature = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
"""TBW."""

import math
import typing as t
from enum import Enum

from typing_extensions import Literal

try:
    import pygmo as pg
except ImportError:
    pass

__all__ = ["Algorithm", "AlgorithmType"]


class AlgorithmType(Enum):
    """TBW."""

    Sade = "sade"
    Sga = "sga"
    Nlopt = "nlopt"


# TODO: Use a class `Sade`, `SGA` and `NLOPT`. See #334
class Algorithm:
    """TBW."""

    def __init__(
        self,
        # TODO: Rename 'type' into 'algorithm_type'. See #334
        type: Literal["sade", "sga", "nlopt"] = "sade",
        generations: int = 1,
        population_size: int = 1,
        # SADE #####
        variant: int = 2,
        variant_adptv: int = 1,
        ftol: float = 1e-6,
        xtol: float = 1e-6,
        memory: bool = False,
        # SGA #####
        cr: float = 0.9,
        eta_c: float = 1.0,
        m: float = 0.02,
        param_m: float = 1.0,
        param_s: int = 2,
        crossover: str = "exponential",
        mutation: str = "polynomial",
        selection: str = "tournament",
        # NLOPT #####
        nlopt_solver: str = "neldermead",
        maxtime: int = 0,
        maxeval: int = 0,
        xtol_rel: float = 1.0e-8,
        xtol_abs: float = 0.0,
        ftol_rel: float = 0.0,
        ftol_abs: float = 0.0,
        stopval: float = -math.inf,
        local_optimizer: t.Optional["pg.nlopt"] = None,
        replacement: str = "best",
        nlopt_selection: str = "best",
    ):
        if generations not in range(1, 100001):
            raise ValueError("'generations' must be between 1 and 100000.")

        if population_size not in range(1, 100001):
            raise ValueError("'population_size' must be between 1 and 100000.")

        self._type = AlgorithmType(type)
        self._generations = generations
        self._population_size = population_size

        # SADE #####
        if variant not in range(1, 19):
            raise ValueError("'variant' must be between 1 and 18.")

        if variant_adptv not in (1, 2):
            raise ValueError("'variant_adptv' must be between 1 and 2.")

        self._variant = variant
        self._variant_adptv = variant_adptv
        self._ftol = ftol
        self._xtol = xtol
        self._memory = memory

        # SGA #####
        if not (0.0 <= cr <= 1.0):
            raise ValueError("'cr' must be between 0.0 and 1.0.")

        if not (0.0 <= m <= 1.0):
            raise ValueError("'m' must be between 0.0 and 1.0.")

        self._cr = cr
        self._eta_c = eta_c
        self._m = m
        self._param_m = param_m
        self._param_s = param_s
        self._crossover = crossover
        self._mutation = mutation
        self._selection = selection

        # NLOPT #####
        self._nlopt_solver = nlopt_solver
        self._maxtime = maxtime
        self._maxeval = maxeval
        self._xtol_rel = xtol_rel
        self._xtol_abs = xtol_abs
        self._ftol_rel = ftol_rel
        self._ftol_abs = ftol_abs
        self._stopval = stopval
        self._local_optimizer = local_optimizer  # type: t.Optional[pg.nlopt]
        self._replacement = replacement
        self._nlopt_selection = nlopt_selection

    @property
    def type(self) -> AlgorithmType:
        """TBW."""
        return self._type

    @type.setter
    def type(self, value: AlgorithmType) -> None:
        """TBW."""
        self._type = AlgorithmType(value)

    @property
    def generations(self) -> int:
        """TBW."""
        return self._generations

    @generations.setter
    def generations(self, value: int) -> None:
        """TBW."""
        if value not in range(1, 100001):
            raise ValueError("'generations' must be between 1 and 100000.")

        self._generations = value

    @property
    def population_size(self) -> int:
        """TBW."""
        return self._population_size

    @population_size.setter
    def population_size(self, value: int) -> None:
        """TBW."""
        if value not in range(1, 100001):
            raise ValueError("'population_size' must be between 1 and 100000.")

        self._population_size = value

    # SADE #####
    @property
    def variant(self) -> int:
        """TBW."""
        return self._variant

    @variant.setter
    def variant(self, value: int) -> None:
        """TBW."""
        if value not in range(1, 19):
            raise ValueError("'variant' must be between 1 and 18.")

        self._variant = value

    @property
    def variant_adptv(self) -> int:
        """TBW."""
        return self._variant_adptv

    @variant_adptv.setter
    def variant_adptv(self, value: int) -> None:
        """TBW."""
        if value not in (1, 2):
            raise ValueError("'variant_adptv' must be between 1 and 2.")

        self._variant_adptv = value

    @property
    def ftol(self) -> float:
        """TBW."""
        return self._ftol

    @ftol.setter
    def ftol(self, value: float) -> None:
        """TBW."""
        self._ftol = value

    @property
    def xtol(self) -> float:
        """TBW."""
        return self._xtol

    @xtol.setter
    def xtol(self, value: float) -> None:
        """TBW."""
        self._xtol = value

    @property
    def memory(self) -> bool:
        """TBW."""
        return self._memory

    @memory.setter
    def memory(self, value: bool) -> None:
        """TBW."""
        self._memory = value

    # SADE #####

    # SGA #####
    @property
    def cr(self) -> float:
        """TBW."""
        return self._cr

    @cr.setter
    def cr(self, value: float) -> None:
        """TBW."""
        if not (0.0 <= value <= 1.0):
            raise ValueError("'cr' must be between 0.0 and 1.0.")

        self._cr = value

    @property
    def eta_c(self) -> float:
        """TBW."""
        return self._eta_c

    @eta_c.setter
    def eta_c(self, value: float) -> None:
        """TBW."""
        self._eta_c = value

    @property
    def m(self) -> float:
        """TBW."""
        return self._m

    @m.setter
    def m(self, value: float) -> None:
        """TBW."""
        if not (0.0 <= value <= 1.0):
            raise ValueError("'m' must be between 0.0 and 1.0.")

        self._m = value

    @property
    def param_m(self) -> float:
        """TBW."""
        return self._param_m

    @param_m.setter
    def param_m(self, value: float) -> None:
        """TBW."""
        self._param_m = value

    @property
    def param_s(self) -> int:
        """TBW."""
        return self._param_s

    @param_s.setter
    def param_s(self, value: int) -> None:
        """TBW."""
        self._param_s = value

    @property
    def crossover(self) -> str:
        """TBW."""
        return self._crossover

    @crossover.setter
    def crossover(self, value: str) -> None:
        """TBW."""
        self._crossover = value

    @property
    def mutation(self) -> str:
        """TBW."""
        return self._mutation

    @mutation.setter
    def mutation(self, value: str) -> None:
        """TBW."""
        self._mutation = value

    @property
    def selection(self) -> str:
        """TBW."""
        return self._selection

    @selection.setter
    def selection(self, value: str) -> None:
        """TBW."""
        self._selection = value

    # SGA #####

    # NLOPT #####
    @property
    def nlopt_solver(self) -> str:
        """TBW."""
        return self._nlopt_solver

    @nlopt_solver.setter
    def nlopt_solver(self, value: str) -> None:
        """TBW."""
        self._nlopt_solver = value

    @property
    def maxtime(self) -> int:
        """TBW."""
        return self._maxtime

    @maxtime.setter
    def maxtime(self, value: int) -> None:
        """TBW."""
        self._maxtime = value

    @property
    def maxeval(self) -> int:
        """TBW."""
        return self._maxeval

    @maxeval.setter
    def maxeval(self, value: int) -> None:
        """TBW."""
        self._maxeval = value

    @property
    def xtol_rel(self) -> float:
        """TBW."""
        return self._xtol_rel

    @xtol_rel.setter
    def xtol_rel(self, value: float) -> None:
        """TBW."""
        self._xtol_rel = value

    @property
    def xtol_abs(self) -> float:
        """TBW."""
        return self._xtol_abs

    @xtol_abs.setter
    def xtol_abs(self, value: float) -> None:
        """TBW."""
        self._xtol_abs = value

    @property
    def ftol_rel(self) -> float:
        """TBW."""
        return self._ftol_rel

    @ftol_rel.setter
    def ftol_rel(self, value: float) -> None:
        """TBW."""
        self._ftol_rel = value

    @property
    def ftol_abs(self) -> float:
        """TBW."""
        return self._ftol_abs

    @ftol_abs.setter
    def ftol_abs(self, value: float) -> None:
        """TBW."""
        self._ftol_abs = value

    @property
    def stopval(self) -> float:
        """TBW."""
        return self._stopval

    @stopval.setter
    def stopval(self, value: float) -> None:
        """TBW."""
        self._stopval = value

    @property
    def local_optimizer(self) -> t.Optional["pg.nlopt"]:
        """TBW."""
        return self._local_optimizer

    @local_optimizer.setter
    def local_optimizer(self, value: t.Optional["pg.nlopt"]) -> None:
        """TBW."""
        self._local_optimizer = value

    @property
    def replacement(self) -> str:
        """TBW."""
        return self._replacement

    @replacement.setter
    def replacement(self, value: str) -> None:
        """TBW."""
        self._replacement = value

    @property
    def nlopt_selection(self) -> str:
        """TBW."""
        return self._nlopt_selection

    @nlopt_selection.setter
    def nlopt_selection(self, value: str) -> None:
        """TBW."""
        self._nlopt_selection = value

    # NLOPT #####

    # TODO: This could be refactored for each if-statement. See #334
    def get_algorithm(self) -> t.Union["pg.sade", "pg.sga", "pg.nlopt"]:
        """TBW.

        :return:
        """
        if self.type is AlgorithmType.Sade:
            sade_algorithm = pg.sade(
                gen=self.generations,
                variant=self.variant,
                variant_adptv=self.variant_adptv,
                ftol=self.ftol,
                xtol=self.xtol,
                memory=self.memory,
            )  # type: pg.sade

            return sade_algorithm

        elif self.type is AlgorithmType.Sga:
            # mutation parameter
            sga_algorithm = pg.sga(
                gen=self.generations,
                cr=self.cr,  # crossover probability
                crossover=self.crossover,  # single, exponential, binomial, sbx
                m=self.m,  # mutation probability
                mutation=self.mutation,  # uniform, gaussian, polynomial
                param_s=self.param_s,  # number of best ind. in 'truncated'/tournament
                selection=self.selection,  # tournament, truncated
                eta_c=self.eta_c,  # distribution index for sbx crossover
                param_m=self.param_m,
            )  # type: pg.sga

            return sga_algorithm

        elif self.type is AlgorithmType.Nlopt:
            opt_algorithm = pg.nlopt(self.nlopt_solver)  # type: pg.nlopt
            opt_algorithm.maxtime = (
                self.maxtime
            )  # stop when the optimization time (in seconds) exceeds maxtime
            opt_algorithm.maxeval = (
                self.maxeval
            )  # stop when the number of function evaluations exceeds maxeval
            opt_algorithm.xtol_rel = self.xtol_rel  # relative stopping criterion for x
            opt_algorithm.xtol_abs = self.xtol_abs  # absolute stopping criterion for x
            opt_algorithm.ftol_rel = self.ftol_rel
            opt_algorithm.ftol_abs = self.ftol_abs
            opt_algorithm.stopval = self.stopval
            opt_algorithm.local_optimizer = self.local_optimizer
            opt_algorithm.replacement = self.replacement
            opt_algorithm.selection = self.nlopt_selection

            return opt_algorithm

        else:
            raise NotImplementedError



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import typing as t

__all__ = ["get_obj_att", "get_obj_by_type", "get_value", "get_state_ids", "copy_state"]


# TODO: Remove this function ? See issue #230.
def get_obj_by_type(obj: t.Any, key: str, obj_type: t.Optional[t.Type] = None) -> t.Any:
    """Get the object associated with the class type following the key chain.

    :param obj:
    :param key:
    :param obj_type:
    :return:
    """
    obj, att = get_obj_att(obj, key, obj_type)
    if obj_type is not None:
        if isinstance(obj, obj_type):
            return obj


# TODO: Remove this function ? See issue #230.
def get_obj_att(
    obj: t.Any, key: str, obj_type: t.Optional[t.Type] = None
) -> t.Tuple[t.Any, str]:
    """Get the object associated with the key.

    Example::

        >>> obj = {"processor": {"pipeline": {"models": [1, 2, 3]}}}
        >>> om.get_obj_att(obj, "processor.pipeline.models")
        ({'models': [1, 2, 3]}, 'models')

    The above example works as well for a user-defined object with a attribute
    objects, i.e. configuration object model.

    :param obj:
    :param key:
    :param obj_type:
    :return: the object and attribute name tuple
    """
    *body, tail = key.split(".")
    for part in body:
        try:
            if isinstance(obj, dict):
                obj = obj[part]
            elif isinstance(obj, list):
                try:
                    index = int(part)
                    obj = obj[index]
                except ValueError:
                    for _, obj_i in enumerate(obj):
                        if hasattr(obj_i, part):
                            obj = getattr(obj_i, part)
                            break
                        elif obj_i.__class__.__name__ == part:
                            if hasattr(obj_i, tail):
                                obj = obj_i
                                break
            elif hasattr(obj, part):
                obj = getattr(obj, part)
            else:
                raise NotImplementedError(
                    f"obj={obj!r}, key={key!r}, obj_type={obj_type!r}, part={part!r}"
                )

            if obj_type and isinstance(obj, obj_type):
                return obj, tail

        except AttributeError:
            # logging.error('Cannot find attribute %r in key %r', part, key)
            obj = None
            break
    return obj, tail


# TODO: Remove this function ? See issue #230.
def get_state(obj: t.Any) -> t.Mapping[str, t.Any]:
    """Convert the config object to a embedded dict object.

    The returned value will be a dictionary tree that is JSON
    compatible. Only dict, list, str, numbers will be contained
    in the returned dictionary.

    :param obj: a top level dict object or a object that defines
        the __getstate__ method.

    :return: the dictionary representation of the passed object.
    """
    result = {}
    if isinstance(obj, dict):
        for key, value in obj.items():
            result[key] = value.__getstate__()

    elif hasattr(obj, "__getstate__"):
        result = obj.__getstate__()

    return result


# TODO: Remove this function ? See issue #230.
def get_state_ids(
    obj: t.Any,
    parent_key_list: t.Optional[t.Sequence[str]] = None,
    result: t.Optional[t.Dict[str, t.Any]] = None,
) -> t.Any:
    """Retrieve a flat dictionary of the object attribute hierarchy.

    The dot-format is used as the key representation.

    :param obj:
    :param parent_key_list:
    :param result:
    :return:
    """
    if result is None:
        from collections import OrderedDict

        result = OrderedDict()

    if parent_key_list is None:
        parent_key_list = []

    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, (str, int, float)):
                key = ".".join(parent_key_list) + "." + key
                result[key] = value
            else:
                list(parent_key_list) + [key]
                get_state_ids(value, list(parent_key_list) + [key], result)

    elif isinstance(obj, list):
        is_primitive = all([isinstance(value, (str, int, float)) for value in obj])
        if is_primitive:
            key = ".".join(parent_key_list)
            result[key] = obj
        else:
            for i, value in enumerate(obj):
                get_state_ids(value, list(parent_key_list) + [str(i)], result)
    return result


# TODO: Remove this function ? See issue #230.
def get_value(obj: t.Any, key: str) -> t.Any:
    """Retrieve the attribute value of the object given the attribute dot formatted key chain.

    Example::

        >>> obj = {"processor": {"pipeline": {"models": [1, 2, 3]}}}
        >>> om.get_value(obj, "processor.pipeline.models")
        [1, 2, 3]

    The above example works as well for a user-defined object with a attribute
    objects, i.e. configuration object model.

    :param obj:
    :param key:
    :return:
    """
    obj, att = get_obj_att(obj, key)

    if isinstance(obj, dict) and att in obj:
        value = obj[att]
    else:
        value = getattr(obj, att)

    return value


# TODO: Remove this function ? See issue #230.
def copy_state(obj: t.Any) -> t.Mapping[str, t.Any]:
    """Deep copy the object as a attribute name/value pairs dictionary.

    :param obj:
    :return:
    """
    kwargs = {}
    for key, _ in obj.__getstate__().items():
        obj_att = getattr(obj, key)
        if obj_att is None:
            cpy_obj = None
        elif hasattr(
            obj_att, "copy"
        ):  # TODO: PYXEL specific: this should be replaced with __copy__
            cpy_obj = obj_att.copy()
        else:
            cpy_obj = type(obj_att)(obj_att)
        kwargs[key] = cpy_obj
    return kwargs



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Charge class to generate electrons or holes inside detector."""

import typing as t

import numba
import numpy as np
import pandas as pd
from typing_extensions import Literal

from pyxel.detectors.geometry import (
    get_horizontal_pixel_center_pos,
    get_vertical_pixel_center_pos,
)

if t.TYPE_CHECKING:
    from pyxel.detectors import Geometry


@numba.jit(nopython=True)
def df_to_array(
    array: np.ndarray,
    charge_per_pixel: list,
    pixel_index_ver: list,
    pixel_index_hor: list,
) -> np.ndarray:
    """Assign charge in dataframe to nearest pixel.

    Parameters
    ----------
    array: ndarray
    charge_per_pixel: list
    pixel_index_ver: list
    pixel_index_hor:list

    Returns
    -------
    ndarray
    """
    for i, charge_value in enumerate(charge_per_pixel):
        array[pixel_index_ver[i], pixel_index_hor[i]] += charge_value
    return array


class Charge:
    """TBW."""

    EXP_TYPE = float
    TYPE_LIST = (
        np.dtype(np.float16),
        np.dtype(np.float32),
        np.dtype(np.float64),
    )

    def __init__(self, geo: "Geometry"):

        self._array = np.zeros(
            (geo.row, geo.col), dtype=self.EXP_TYPE
        )  # type: np.ndarray
        self._geo = geo
        self.nextid = 0  # type: int

        self.columns = (
            "charge",
            "number",
            "init_energy",
            "energy",
            "init_pos_ver",
            "init_pos_hor",
            "init_pos_z",
            "position_ver",
            "position_hor",
            "position_z",
            "velocity_ver",
            "velocity_hor",
            "velocity_z",
        )  # type: t.Tuple[str, ...]

        self.EMPTY_FRAME = pd.DataFrame(
            columns=self.columns, dtype=float
        )  # type: pd.DataFrame
        self._frame = self.EMPTY_FRAME.copy()  # type: pd.DataFrame

    @staticmethod
    def create_charges(
        *,
        particle_type: Literal["e", "h"],
        particles_per_cluster: np.ndarray,
        init_energy: np.ndarray,
        init_ver_position: np.ndarray,
        init_hor_position: np.ndarray,
        init_z_position: np.ndarray,
        init_ver_velocity: np.ndarray,
        init_hor_velocity: np.ndarray,
        init_z_velocity: np.ndarray,
    ) -> pd.DataFrame:
        """Create new charge(s) or group of charge(s) as a `DataFrame`.

        Parameters
        ----------
        particle_type : str
            Type of particle. Valid values: 'e' for an electron or 'h' for a hole.
        particles_per_cluster : array-like
        init_energy : array
        init_ver_position : array
        init_hor_position : array
        init_z_position : array
        init_ver_velocity : array
        init_hor_velocity : array
        init_z_velocity : array

        Returns
        -------
        DataFrame
            Charge(s) stored in a ``DataFrame``.
        """
        if not (
            len(particles_per_cluster)
            == len(init_energy)
            == len(init_ver_position)
            == len(init_hor_position)
            == len(init_z_position)
            == len(init_ver_velocity)
            == len(init_hor_velocity)
            == len(init_z_velocity)
        ):
            raise ValueError("List arguments have different lengths.")

        if not (
            particles_per_cluster.ndim
            == init_energy.ndim
            == init_ver_position.ndim
            == init_hor_position.ndim
            == init_z_position.ndim
            == init_ver_velocity.ndim
            == init_hor_velocity.ndim
            == init_z_velocity.ndim
            == 1
        ):
            raise ValueError("List arguments must have only one dimension.")

        elements = len(init_energy)

        # check_position(self.detector, init_ver_position, init_hor_position, init_z_position)      # TODO
        # check_energy(init_energy)             # TODO
        # Check if particle number is integer:
        # check_type(particles_per_cluster)      # TODO

        if particle_type == "e":
            charge = [-1] * elements  # * cds.e
        elif particle_type == "h":
            charge = [+1] * elements  # * cds.e
        else:
            raise ValueError("Given charged particle type can not be simulated")

        # if all(init_ver_velocity) == 0 and all(init_hor_velocity) == 0 and all(init_z_velocity) == 0:
        #     random_direction(1.0)

        # Create new charges as a `dict`
        new_charges = {
            "charge": charge,
            "number": particles_per_cluster,
            "init_energy": init_energy,
            "energy": init_energy,
            "init_pos_ver": init_ver_position,
            "init_pos_hor": init_hor_position,
            "init_pos_z": init_z_position,
            "position_ver": init_ver_position,
            "position_hor": init_hor_position,
            "position_z": init_z_position,
            "velocity_ver": init_ver_velocity,
            "velocity_hor": init_hor_velocity,
            "velocity_z": init_z_velocity,
        }  # type: t.Mapping[str, t.Union[t.Sequence, np.ndarray]]

        return pd.DataFrame(new_charges)

    def convert_df_to_array(self):
        """Convert charge dataframe to an array.

        Charge in the detector volume is collected and assigned to the nearest pixel.
        """
        array = np.zeros((self._geo.row, self._geo.col))

        charge_per_pixel = self.get_frame_values(quantity="number")
        charge_pos_ver = self.get_frame_values(quantity="position_ver")
        charge_pos_hor = self.get_frame_values(quantity="position_hor")

        pixel_index_ver = np.floor_divide(
            charge_pos_ver, self._geo.pixel_vert_size
        ).astype(int)
        pixel_index_hor = np.floor_divide(
            charge_pos_hor, self._geo.pixel_horz_size
        ).astype(int)

        # Changing = to += since charge dataframe is reset, the pixel array need to be
        # incremented, we can't do the whole operation on each iteration
        return df_to_array(array, charge_per_pixel, pixel_index_ver, pixel_index_hor)

    @staticmethod
    def convert_array_to_df(
        array: np.ndarray,
        num_rows: int,
        num_cols: int,
        pixel_vertical_size: float,
        pixel_horizontal_size: float,
    ) -> pd.DataFrame:
        """Convert charge array to a dataframe placing charge packets in pixels to the center and on top of pixels.

        Parameters
        ----------
        array: ndarray
        num_rows: int
        num_cols: int
        pixel_vertical_size: float
        pixel_horizontal_size: float

        Returns
        -------
        DataFrame
        """
        charge_number = array.flatten()
        where_non_zero = np.where(charge_number > 0.0)
        charge_number = charge_number[where_non_zero]
        size = charge_number.size  # type: int

        vertical_pixel_center_pos_1d = get_vertical_pixel_center_pos(
            num_rows=num_rows,
            num_cols=num_cols,
            pixel_vertical_size=pixel_vertical_size,
        )

        horizontal_pixel_center_pos_1d = get_horizontal_pixel_center_pos(
            num_rows=num_rows,
            num_cols=num_cols,
            pixel_horizontal_size=pixel_horizontal_size,
        )

        init_ver_pix_position_1d = vertical_pixel_center_pos_1d[where_non_zero]
        init_hor_pix_position_1d = horizontal_pixel_center_pos_1d[where_non_zero]

        # Create new charges
        return Charge.create_charges(
            particle_type="e",
            particles_per_cluster=charge_number,
            init_energy=np.zeros(size),
            init_ver_position=init_ver_pix_position_1d,
            init_hor_position=init_hor_pix_position_1d,
            init_z_position=np.zeros(size),
            init_ver_velocity=np.zeros(size),
            init_hor_velocity=np.zeros(size),
            init_z_velocity=np.zeros(size),
        )

    def add_charge_dataframe(self, new_charges: pd.DataFrame) -> None:
        """Add new charge(s) or group of charge(s) to the charge dataframe.

        Parameters
        ----------
        new_charges : DataFrame
            Charges as a `DataFrame`
        """
        if set(new_charges.columns) != set(self.columns):
            expected_columns = ", ".join(map(repr, self.columns))  # type: str
            raise ValueError(f"Expected columns: {expected_columns}")

        if self._frame.empty:
            if not np.all(self.array == 0):
                df = Charge.convert_array_to_df(
                    array=self.array,
                    num_cols=self._geo.col,
                    num_rows=self._geo.row,
                    pixel_vertical_size=self._geo.pixel_vert_size,
                    pixel_horizontal_size=self._geo.pixel_horz_size,
                )
                new_frame = df.append(new_charges, ignore_index=True)
            else:
                new_frame = new_charges
        else:
            new_frame = self._frame.append(new_charges, ignore_index=True)

        self._frame = new_frame
        self.nextid = self.nextid + len(new_charges)

    def add_charge(
        self,
        *,
        particle_type: Literal["e", "h"],
        particles_per_cluster: np.ndarray,
        init_energy: np.ndarray,
        init_ver_position: np.ndarray,
        init_hor_position: np.ndarray,
        init_z_position: np.ndarray,
        init_ver_velocity: np.ndarray,
        init_hor_velocity: np.ndarray,
        init_z_velocity: np.ndarray,
    ) -> None:
        """Add new charge(s) or group of charge(s) inside the detector.

        Parameters
        ----------
        particle_type : str
            Type of particle. Valid values: 'e' for an electron or 'h' for a hole.
        particles_per_cluster : array
        init_energy : array
        init_ver_position : array
        init_hor_position : array
        init_z_position : array
        init_ver_velocity : array
        init_hor_velocity : array
        init_z_velocity : array
        """
        # Create charge(s)
        new_charges = Charge.create_charges(
            particle_type=particle_type,
            particles_per_cluster=particles_per_cluster,
            init_energy=init_energy,
            init_ver_position=init_ver_position,
            init_hor_position=init_hor_position,
            init_z_position=init_z_position,
            init_ver_velocity=init_ver_velocity,
            init_hor_velocity=init_hor_velocity,
            init_z_velocity=init_z_velocity,
        )  # type: pd.DataFrame

        # Add charge(s)
        self.add_charge_dataframe(new_charges=new_charges)

    def add_charge_array(self, array: np.ndarray) -> None:
        """Add charge to the charge array. Add to charge dataframe if not empty instead.

        Parameters
        ----------
        array: ndarray
        """
        self.validate_type(array)
        self.validate_shape(array)

        if self._frame.empty:
            self._array += array

        else:
            charge_df = Charge.convert_array_to_df(
                array=array,
                num_cols=self._geo.col,
                num_rows=self._geo.row,
                pixel_vertical_size=self._geo.pixel_vert_size,
                pixel_horizontal_size=self._geo.pixel_horz_size,
            )
            self.add_charge_dataframe(charge_df)

    @property
    def array(self) -> np.ndarray:
        """Get charge in a numpy array."""
        if not self._frame.empty:
            self._array = self.convert_df_to_array()
        return self._array

    def __array__(self, dtype: t.Optional[np.dtype] = None):
        if not isinstance(self._array, np.ndarray):
            raise TypeError("Array not initialized.")
        return np.asarray(self._array, dtype=dtype)

    @property
    def frame(self) -> pd.DataFrame:
        """Get charge in a pandas dataframe."""
        if not isinstance(self._frame, pd.DataFrame):
            raise TypeError("Charge data frame not initialized.")
        return self._frame

    def empty(self) -> None:
        """Empty all data stored in Charge class."""
        self.nextid = 0
        if not self._frame.empty:
            self._frame = self.EMPTY_FRAME.copy()
        self._array *= 0

    def validate_type(self, value: np.ndarray) -> None:
        """Validate a value.

        Parameters
        ----------
        value
        """
        cls_name = self.__class__.__name__  # type: str

        if not isinstance(value, np.ndarray):
            raise TypeError(f"{cls_name} array should be a numpy.ndarray")

        if value.dtype not in self.TYPE_LIST:
            exp_type_name = str(self.EXP_TYPE)  # type: str
            raise TypeError(f"Expected type of {cls_name} array is {exp_type_name}.")

    def validate_shape(self, value: np.ndarray) -> None:
        """TBW."""
        cls_name = self.__class__.__name__  # type: str

        if value.shape != self._array.shape:
            raise ValueError(f"Expected {cls_name} array is {self._array.shape}.")

    def get_frame_values(
        self, quantity: str, id_list: t.Optional[list] = None
    ) -> np.ndarray:
        """Get quantity values of particles defined with id_list. By default it returns values of all particles.

        Parameters
        ----------
        quantity : str
            Name of the quantity: ``number``, ``energy``, ``position_ver``, ``velocity_hor``, etc.
        id_list : Sequence of int
            List of particle ids: ``[0, 12, 321]``

        Returns
        -------
        array
        """
        if id_list:
            df = self._frame.query("index in %s" % id_list)  # type: pd.DataFrame
        else:
            df = self._frame

        array = df[quantity].values  # type: np.ndarray

        return array

    def set_frame_values(
        self, quantity: str, new_value_list: list, id_list: t.Optional[list] = None
    ) -> None:
        """Update quantity values of particles defined with id_list. By default it updates all.

        Parameters
        ----------
        quantity : str
            Name of the quantity: ``number``, ``energy``, ``position_ver``, ``velocity_hor``, etc.
        new_value_list : Sequence of int
            List of values ``[1.12, 2.23, 3.65]``
        id_list : Sequence of int
            List of particle ids: ``[0, 12, 321]``
        """
        new_df = pd.DataFrame({quantity: new_value_list}, index=id_list)
        self._frame.update(new_df)

    def remove_from_frame(self, id_list: t.Optional[list] = None) -> None:
        """Remove particles defined with id_list. By default it removes all particles from DataFrame.

        Parameters
        ----------
        id_list : Sequence of int
            List of particle ids: ``[0, 12, 321]``
        """
        if id_list:
            # TODO: Check carefully if 'inplace' is needed. This could break lot of things.
            self._frame.query("index not in %s" % id_list, inplace=True)
        else:
            self._frame = self.EMPTY_FRAME.copy()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Persistence classes."""
import typing as t
from dataclasses import dataclass

import numpy as np


@dataclass
class Trap:
    """Trap dataclass."""

    time_constant: float
    proportion: float
    charge: np.ndarray


class Persistence:
    """Persistence class."""

    def __init__(
        self,
        trap_time_constants: t.Sequence[float],
        trap_proportions: t.Sequence[float],
        geometry: t.Tuple[int, ...],
    ):
        traps = []
        for time_constant, trap_proportion in sorted(
            zip(trap_time_constants, trap_proportions)
        ):
            traps.append(
                Trap(
                    time_constant=time_constant,
                    proportion=trap_proportion,
                    charge=np.zeros(geometry),
                )
            )
        self._trap_list = traps
        self._trapped_charge_array = np.zeros((len(traps), geometry[0], geometry[1]))
        self._trap_time_constants = np.array(trap_time_constants)
        self._trap_proportions = np.array(trap_proportions)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}<number_of_traps={len(self._trap_list)}>"

    @property
    def trap_list(self) -> t.Sequence[Trap]:
        """List of traps."""
        return self._trap_list

    @property
    def trapped_charge_array(self) -> np.ndarray:
        """Return an array of trapped charge."""
        # for i, trap in enumerate(self.trap_list):
        #     out[i] = trap.charge
        return self._trapped_charge_array

    @trapped_charge_array.setter
    def trapped_charge_array(self, charge_3d: np.ndarray) -> None:
        """Set an array of trapped charge."""
        for i, trap in enumerate(self.trap_list):
            if not charge_3d[i].shape == trap.charge.shape:
                raise ValueError(
                    "Mismatch in shapes between saved trapped charge and input charge."
                )
            trap.charge = charge_3d[i]
        self._trapped_charge_array = charge_3d

    @property
    def trap_time_constants(self) -> np.ndarray:
        """Return trap time constants."""
        return self._trap_time_constants

    @property
    def trap_proportions(self) -> np.ndarray:
        """Return trap proportions."""
        return self._trap_proportions


@dataclass
class SimpleTrap:
    """Simple trap dataclass."""

    time_constant: float
    density: float
    charge: np.ndarray


class SimplePersistence:
    """Simple persistence dataclass."""

    def __init__(
        self,
        trap_time_constants: t.Sequence[float],
        trap_densities: t.Sequence[float],
        geometry: t.Tuple[int, ...],
    ):
        traps = []
        for time_constant, trap_density in sorted(
            zip(trap_time_constants, trap_densities)
        ):
            traps.append(
                SimpleTrap(
                    time_constant=time_constant,
                    density=trap_density,
                    charge=np.zeros(geometry),
                )
            )
        self._trap_list = traps
        self._trapped_charge_array = np.zeros((len(traps), geometry[0], geometry[1]))
        self._trap_time_constants = np.array(trap_time_constants)
        self._trap_densities = np.array(trap_densities)

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}<number_of_traps={len(self._trap_list)}>"

    @property
    def trap_list(self) -> t.Sequence[SimpleTrap]:
        """Return a list of traps."""
        return self._trap_list

    @property
    def trapped_charge_array(self) -> np.ndarray:
        """Return an array of trapped charge."""
        # for i, trap in enumerate(self.trap_list):
        #     out[i] = trap.charge
        return self._trapped_charge_array

    @trapped_charge_array.setter
    def trapped_charge_array(self, charge_3d: np.ndarray) -> None:
        """Set an array of trapped charge."""
        for i, trap in enumerate(self.trap_list):
            if not charge_3d[i].shape == trap.charge.shape:
                raise ValueError(
                    "Mismatch in shapes between saved trapped charge and input charge."
                )
            trap.charge = charge_3d[i]
        self._trapped_charge_array = charge_3d

    @property
    def trap_time_constants(self) -> np.ndarray:
        """Return trap time constants."""
        return self._trap_time_constants

    @property
    def trap_densities(self) -> np.ndarray:
        """Return trap densities."""
        return self._trap_densities



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Timing functions."""

import timeit
import typing as t

import numpy as np
import pandas as pd

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector
    from pyxel.pipelines import DetectionPipeline, ModelFunction, ModelGroup


def _list_of_times_to_dataframe(times: list, model_names: list) -> pd.DataFrame:
    """Convert a list of model times to a dataframe.

    Parameters
    ----------
    times: list
    model_names: list

    Returns
    -------
    df: pd.DataFrame
    """

    times_array = np.array(times) * 1000  # to miliseconds
    times_sum = np.sum(times_array)
    percentages = times_array * 100 / times_sum

    final_times = np.round(np.append(times_array, times_sum), 2)
    percentages = np.round(np.append(percentages, 100.0), 4)

    model_names.append("TOTAL:")

    df = pd.DataFrame({"time [ms]": final_times, "%time": percentages})
    df.index = model_names

    return df


def time_pipeline(detector: "Detector", pipeline: "DetectionPipeline") -> pd.DataFrame:
    """Time a single pipeline.

    Parameters
    ----------
    detector : Detector
    pipeline : DetectionPipeline

    Returns
    -------
    df: DataFrame
    """

    times = []
    model_names = []

    for group_name in pipeline.model_group_names:

        models_grp = getattr(pipeline, group_name)  # type: t.Optional[ModelGroup]

        if models_grp:
            for model in models_grp.models:  # type: ModelFunction
                if model.enabled:
                    model_start = timeit.default_timer()
                    model(detector)
                    model_end = timeit.default_timer()
                    times.append(model_end - model_start)
                    model_names.append(model.name)

    df = _list_of_times_to_dataframe(times, model_names)

    return df



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Fitness functions for model fitting."""

import numba
import numpy as np


@numba.njit
def sum_of_abs_residuals(
    simulated: np.ndarray,
    target: np.ndarray,
    weighting: np.ndarray,
) -> float:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    diff = target - simulated
    diff *= weighting

    result = float(np.sum(np.abs(diff)))
    return result


@numba.njit
def sum_of_squared_residuals(
    simulated: np.ndarray,
    target: np.ndarray,
    weighting: np.ndarray,
) -> float:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    diff = target - simulated
    diff_square = diff * diff
    diff_square *= weighting

    result = float(np.sum(diff_square))
    return result



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Charge readout model."""

import typing as t

import numpy as np

from pyxel.detectors import Detector


def apply_gain(pixel_2d: np.ndarray, gain: float) -> np.ndarray:
    """Apply a gain (in V/e-) to a pixel array (in e-).

    Parameters
    ----------
    pixel_2d : ndarray
        2D array of pixels. Unit: e-
    gain : float
        Gain to apply. Unit: V/e-

    Returns
    -------
    ndarray
        2D array of signals. Unit: V
    """
    new_data_2d = pixel_2d * gain

    return new_data_2d


def simple_measurement(detector: Detector, gain: t.Optional[float] = None) -> None:
    """Convert the pixel array into signal array.

    Notes
    -----
    If no gain is provided, then its value will be the sensitivity of charge readout
    provided in the ``Detector`` object.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    gain : float, optional.
        Gain to apply. By default, this is the sensitivity of charge readout. Unit: V/e-
    """
    if gain is None:
        char = detector.characteristics
        gain = char.charge_to_volt_conversion

    # Compute
    signal_2d = apply_gain(pixel_2d=detector.pixel.array, gain=gain)

    detector.signal.array = signal_2d.astype("float64")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
import typing as t  # noqa: F401
from copy import deepcopy

from pyxel import util
from pyxel.pipelines import ModelFunction

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector
    from pyxel.pipelines import DetectionPipeline


# TODO: These methods could also be as a `abc.Sequence` with magical methods:
#       __getitem__, __iter__, __len__, __contains__, ...
#       See #181
class ModelGroup:
    """TBW."""

    def __init__(self, models: t.Sequence[ModelFunction], name: str):
        self._log = logging.getLogger(__name__)
        self._name = name
        self.models = models  # type: t.Sequence[ModelFunction]

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str

        all_models = [
            model.name for model in self.models if model.name
        ]  # type: t.List[str]

        return f"{cls_name}<name={self._name!r}, models={all_models!r}>"

    def __deepcopy__(self, memo: dict) -> "ModelGroup":
        copied_models = deepcopy(self.models)
        return ModelGroup(models=copied_models, name=self._name)

    def __iter__(self) -> t.Iterator[ModelFunction]:
        for model in self.models:
            if model.enabled:
                yield model

    def __getstate__(self) -> t.Mapping:
        return {"models": tuple(self.models), "name": self._name}

    def __setstate__(self, state: t.Mapping) -> None:
        self.models = list(state["models"])
        self._name = state["name"]

    def __getattr__(self, item: str) -> ModelFunction:
        for model in self.models:
            if model.name == item:
                return model
        else:
            raise AttributeError(f"Cannot find model {item!r}.")

    def __dir__(self):
        return dir(type(self)) + [model.name for model in self.models]

    # TODO: Why is this method returning a `bool` ?
    def run(
        self,
        detector: "Detector",
        pipeline: "DetectionPipeline",
        abort_model: t.Optional[str] = None,
    ) -> bool:
        """Execute each enabled model in this group.

        Parameters
        ----------
        detector : Detector
        pipeline : DetectionPipeline
        abort_model : str, optional

        Returns
        -------
        bool
            TBW.
        """
        for model in self.models:  # type: ModelFunction
            if model.name == abort_model:
                return True
            if not pipeline.is_running:
                raise util.PipelineAborted("Pipeline has been aborted.")
            else:
                if model.enabled:
                    self._log.info("Model: %r", model.name)
                    model(detector)

        return False
